#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Objects defined in the Scripting API specification represented as classes that are basically dict-wrappers.

.. autosummary::
    :toctree: _dictionaries

    wotpy.wot.dictionaries.base
    wotpy.wot.dictionaries.filter
    wotpy.wot.dictionaries.interaction
    wotpy.wot.dictionaries.link
    wotpy.wot.dictionaries.schema
    wotpy.wot.dictionaries.security
    wotpy.wot.dictionaries.thing
    wotpy.wot.dictionaries.version
"""



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Classes that represent the JSON and JSON-LD serialization formats of a Thing Description document.
"""

import json

import jsonschema
import six

from wotpy.wot.dictionaries.thing import ThingFragment
from wotpy.wot.thing import Thing
from wotpy.wot.validation import SCHEMA_THING, InvalidDescription


class ThingDescription(object):
    """Class that represents a Thing Description document.
    Contains logic to validate and transform a Thing to a serialized TD and vice versa."""

    def __init__(self, doc):
        """Constructor.
        Validates that the document conforms to the TD schema."""

        self._doc = json.loads(doc) if isinstance(doc, (six.string_types, bytes)) else doc
        self._thing_fragment = ThingFragment(self._doc)

        self.validate(doc=self._thing_fragment.to_dict())

    @classmethod
    def validate(cls, doc):
        """Validates the given Thing Description document against its schema.
        Raises ValidationError if validation fails."""

        try:
            jsonschema.validate(doc, SCHEMA_THING)
        except (jsonschema.ValidationError, TypeError) as ex:
            raise InvalidDescription(str(ex))

    @classmethod
    def from_thing(cls, thing):
        """Builds an instance of a JSON-serialized Thing Description from a Thing object."""

        return ThingDescription(thing.thing_fragment.to_dict())

    def __getattr__(self, name):
        """Search for members that raised an AttributeError in
        the internal ThingFragment before propagating the exception."""

        return getattr(self._thing_fragment, name)

    def to_dict(self):
        """Returns the JSON Thing Description as a dict."""

        return self._thing_fragment.to_dict()

    def to_str(self):
        """Returns the JSON Thing Description as a string."""

        return json.dumps(self._thing_fragment.to_dict())

    def to_thing_fragment(self):
        """Returns a ThingFragment dictionary built from this TD."""

        return self._thing_fragment

    def build_thing(self):
        """Builds a new Thing object from the serialized Thing Description."""

        return Thing(thing_fragment=self.to_thing_fragment())

    def get_forms(self, name):
        """Returns a list of FormDict for the interaction that matches the given name."""

        if name in self.properties:
            return self.get_property_forms(name)

        if name in self.actions:
            return self.get_action_forms(name)

        if name in self.events:
            return self.get_event_forms(name)

        return []

    def get_property_forms(self, name):
        """Returns a list of FormDict for the property that matches the given name."""

        return self.properties[name].forms

    def get_action_forms(self, name):
        """Returns a list of FormDict for the action that matches the given name."""

        return self.actions[name].forms

    def get_event_forms(self, name):
        """Returns a list of FormDict for the event that matches the given name."""

        return self.events[name].forms



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
CoAP resources to deal with Event interactions.
"""

import json
import logging
import time

import aiocoap
import aiocoap.error
import aiocoap.resource
import tornado.gen

from wotpy.protocols.coap.resources.utils import parse_request_opt_query

JSON_CONTENT_FORMAT = 50


def get_thing_event(server, request):
    """Takes a CoAP request and returns the Thing Event
    identified by the request arguments."""

    query = parse_request_opt_query(request)
    url_name_thing = query.get("thing")
    url_name_event = query.get("name")

    if not url_name_thing or not url_name_event:
        raise aiocoap.error.BadRequest("Missing query arguments")

    exposed_thing = server.exposed_thing_set.find_by_thing_id(url_name_thing)

    if not exposed_thing:
        raise aiocoap.error.NotFound("Thing not found")

    try:
        return next(
            exposed_thing.events[key] for key in exposed_thing.events
            if exposed_thing.events[key].url_name == url_name_event)
    except StopIteration:
        raise aiocoap.error.NotFound("Event not found")


class EventResource(aiocoap.resource.ObservableResource):
    """CoAP resource to observe Event emissions."""

    def __init__(self, server):
        super(EventResource, self).__init__()
        self._server = server
        self._subscription = None
        self._last_events = {}
        self._logr = logging.getLogger(__name__)

    @classmethod
    def _event_key(cls, thing_event):
        """Returns the internal event key for the given Thing Event."""

        return thing_event.thing.url_name, thing_event.url_name

    @tornado.gen.coroutine
    def add_observation(self, request, server_observation):
        """Method that decides whether to add a new observer.
        A new observer is added for each GET request."""

        if request.code.name != aiocoap.Code.GET.name:
            return

        try:
            thing_event = get_thing_event(self._server, request)
        except aiocoap.error.Error:
            return

        def on_next(item):
            event_item = {
                "name": item.name,
                "data": item.data,
                "time": int(time.time() * 1000)
            }

            self._last_events[self._event_key(thing_event)] = event_item
            server_observation.trigger()

        def on_error(err):
            self._logr.warning("Error on subscription to {}: {}".format(thing_event, err))

        subscription = thing_event.subscribe(on_next=on_next, on_error=on_error)

        def cancellation_cb():
            self._logr.debug("Disposing of subscription to: {}".format(thing_event))
            subscription.dispose()

        server_observation.accept(cancellation_cb)

    @tornado.gen.coroutine
    def render_get(self, request):
        """Returns a CoAP response with the last observed event emission."""

        thing_event = get_thing_event(self._server, request)
        last_item = self._last_events.get(self._event_key(thing_event), None)
        payload = json.dumps(last_item).encode("utf-8") if last_item else b""
        response = aiocoap.Message(code=aiocoap.Code.CONTENT, payload=payload)
        response.opt.content_format = JSON_CONTENT_FORMAT

        raise tornado.gen.Return(response)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Enumeration classes related to the WebSockets server.
"""

from wotpy.utils.enums import EnumListMixin


class WebsocketMethods(EnumListMixin):
    """Enumeration of available websocket message actions."""

    READ_PROPERTY = "read_property"
    WRITE_PROPERTY = "write_property"
    INVOKE_ACTION = "invoke_action"
    ON_PROPERTY_CHANGE = "on_property_change"
    ON_TD_CHANGE = "on_td_change"
    ON_EVENT = "on_event"
    DISPOSE = "dispose"


class WebsocketErrors(EnumListMixin):
    """Enumeration of JSON RPC error codes."""

    PARSE_ERROR = -32700
    INVALID_REQUEST = -32600
    METHOD_NOT_FOUND = -32601
    INVALID_METHOD_PARAMS = -32602
    INTERNAL_ERROR = -32603
    SUBSCRIPTION_ERROR = -32000


class WebsocketSchemes(EnumListMixin):
    """Enumeration of Websocket schemes."""

    WS = "ws"
    WSS = "wss"



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import socket

import six
import tornado.ioloop
from tornado.escape import to_unicode

DEFAULT_TIMEOUT_SECS = 30
TIMEOUT_CORO_VAR = "WOTPY_TESTS_CORO_TIMEOUT"


def run_test_coroutine(coro, timeout=None):
    """Synchronously runs the given test coroutine with an optinally defined timeout."""

    timeout = timeout if timeout else os.getenv(TIMEOUT_CORO_VAR, DEFAULT_TIMEOUT_SECS)

    tornado.ioloop.IOLoop.current().run_sync(coro, timeout=float(timeout))


def assert_equal_dict(dict_a, dict_b, compare_as_unicode=False):
    """Asserts that both dicts are equal."""

    assert set(dict_a.keys()) == set(dict_b.keys())

    for key in dict_a:
        value_a = dict_a[key]
        value_b = dict_b[key]

        if compare_as_unicode and isinstance(value_a, six.string_types):
            assert to_unicode(value_a) == to_unicode(value_b)
        else:
            assert value_a == value_b


def find_free_port():
    """Returns a free TCP port by attempting to open a socket on an OS-assigned port."""

    try:
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        sock.bind(("", 0))
        return sock.getsockname()[1]
    finally:
        # noinspection PyUnboundLocalVariable
        sock.close()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Class that handles incoming WebSockets messages.
"""

import uuid

from jsonschema import validate, ValidationError
from rx.concurrency import IOLoopScheduler
from tornado import websocket, gen

from wotpy.protocols.ws.enums import WebsocketMethods, WebsocketErrors
from wotpy.protocols.ws.messages import \
    WebsocketMessageRequest, \
    WebsocketMessageException, \
    WebsocketMessageError, \
    WebsocketMessageResponse, \
    WebsocketMessageEmittedItem
from wotpy.protocols.ws.schemas import \
    SCHEMA_PARAMS_READ_PROPERTY, \
    SCHEMA_PARAMS_WRITE_PROPERTY, \
    SCHEMA_PARAMS_DISPOSE, \
    SCHEMA_PARAMS_INVOKE_ACTION, \
    SCHEMA_PARAMS_ON_PROPERTY_CHANGE, \
    SCHEMA_PARAMS_ON_TD_CHANGE, \
    SCHEMA_PARAMS_ON_EVENT


# noinspection PyAbstractClass
class WebsocketHandler(websocket.WebSocketHandler):
    """Tornado handler for Websocket messages.
    This class processes all incoming WebSocket messages and
    translates them to actions executed on ExposedThing objects."""

    POLICY_VIOLATION_CODE = 1008
    POLICY_VIOLATION_REASON = "Not found"

    def __init__(self, *args, **kwargs):
        self._server = kwargs.pop("websocket_server", None)
        self._scheduler = IOLoopScheduler()
        self._subscriptions = {}
        self._exposed_thing_name = None
        super(WebsocketHandler, self).__init__(*args, **kwargs)

    @property
    def exposed_thing(self):
        """Exposed thing property.
        Retrieves the ExposedThing from the parent server."""

        try:
            return self._server.get_exposed_thing(self._exposed_thing_name)
        except ValueError:
            self.close(self.POLICY_VIOLATION_CODE, self.POLICY_VIOLATION_REASON)

    def check_origin(self, origin):
        """Should return True to accept the request or False to reject it.
        The origin argument is the value of the Origin HTTP header,
        the url responsible for initiating this request"""

        # ToDo: Check this once we add authentication
        # This is extremely dangerous in case of a cookie-based authentication system.
        # WS authentication should be handled independently with some kind of token-based system.

        return True

    def open(self, name):
        """Called when the WebSockets connection is opened."""

        assert self._exposed_thing_name is None

        try:
            self._server.get_exposed_thing(name)
            self._exposed_thing_name = name
        except ValueError:
            self.close(self.POLICY_VIOLATION_CODE, self.POLICY_VIOLATION_REASON)

    def _write_error(self, message, code, msg_id=None, data=None):
        """Builds an error message instance and sends it to the client."""

        err = WebsocketMessageError(message=message, code=code, data=data, msg_id=msg_id)
        self.write_message(err.to_json())

    def _dispose_subscription(self, subscription_id):
        """Takes a subscription ID and destroys the related subscription."""

        if subscription_id in self._subscriptions:
            subscription = self._subscriptions.pop(subscription_id)
            subscription.dispose()

    def _on_subscription_error(self, subscription_id, err):
        """Default error callback for Observable subscriptions."""

        self._dispose_subscription(subscription_id)
        data_err = {"subscription": subscription_id}
        self._write_error(str(err), WebsocketErrors.SUBSCRIPTION_ERROR, data=data_err)

    def _on_subscription_next(self, subscription_id, item):
        """Default next callback for Observable subscriptions."""

        try:
            msg = WebsocketMessageEmittedItem(
                subscription_id=subscription_id,
                name=item.name,
                data=item.data)
            self.write_message(msg.to_json())
        except WebsocketMessageException as ex:
            self._on_subscription_error(subscription_id, ex)

    def _on_subscription_completed(self, subscription_id):
        """Default completed callback for Observable subscriptions."""

        self._dispose_subscription(subscription_id)

    def _subscribe(self, subscription_id, observable):
        """Subscribe to the given Observable and add the subscription handler to the internal dict."""

        subscription = observable.observe_on(self._scheduler).subscribe(
            on_next=lambda item: self._on_subscription_next(subscription_id, item),
            on_error=lambda err: self._on_subscription_error(subscription_id, err),
            on_completed=lambda: self._on_subscription_completed(subscription_id))

        self._subscriptions[subscription_id] = subscription

    @gen.coroutine
    def _handle_get_property(self, req):
        """Handler for the 'get_property' method."""

        params = req.params

        try:
            validate(params, SCHEMA_PARAMS_READ_PROPERTY)
        except ValidationError as ex:
            self._write_error(str(ex), WebsocketErrors.INVALID_METHOD_PARAMS, msg_id=req.id)
            return

        try:
            prop_value = yield self.exposed_thing.read_property(name=params["name"])
        except Exception as ex:
            self._write_error(str(ex), WebsocketErrors.INTERNAL_ERROR, msg_id=req.id)
            return

        res = WebsocketMessageResponse(result=prop_value, msg_id=req.id)
        self.write_message(res.to_json())

    @gen.coroutine
    def _handle_set_property(self, req):
        """Handler for the 'set_property' method."""

        params = req.params

        try:
            validate(params, SCHEMA_PARAMS_WRITE_PROPERTY)
        except ValidationError as ex:
            self._write_error(str(ex), WebsocketErrors.INVALID_METHOD_PARAMS, msg_id=req.id)
            return

        try:
            yield self.exposed_thing.write_property(name=params["name"], value=params["value"])
        except Exception as ex:
            self._write_error(str(ex), WebsocketErrors.INTERNAL_ERROR, msg_id=req.id)
            return

        res = WebsocketMessageResponse(result=None, msg_id=req.id)
        self.write_message(res.to_json())

    @gen.coroutine
    def _handle_invoke_action(self, req):
        """Handler for the 'invoke_action' method."""

        params = req.params

        try:
            validate(params, SCHEMA_PARAMS_INVOKE_ACTION)
        except ValidationError as ex:
            self._write_error(str(ex), WebsocketErrors.INVALID_METHOD_PARAMS, msg_id=req.id)
            return

        try:
            input_value = params.get("parameters")
            action_result = yield self.exposed_thing.invoke_action(params["name"], input_value)
        except Exception as ex:
            self._write_error(str(ex), WebsocketErrors.INTERNAL_ERROR, msg_id=req.id)
            return

        res = WebsocketMessageResponse(result=action_result, msg_id=req.id)
        self.write_message(res.to_json())

    @gen.coroutine
    def _handle_on_property_change(self, req):
        """Handler for the 'on_property_change' subscription method."""

        params = req.params

        try:
            validate(params, SCHEMA_PARAMS_ON_PROPERTY_CHANGE)
        except ValidationError as ex:
            self._write_error(str(ex), WebsocketErrors.INVALID_METHOD_PARAMS, msg_id=req.id)
            return

        subscription_id = str(uuid.uuid4())

        res = WebsocketMessageResponse(result=subscription_id, msg_id=req.id)
        self.write_message(res.to_json())

        observable = self.exposed_thing.on_property_change(name=params["name"])

        self._subscribe(subscription_id, observable)

    @gen.coroutine
    def _handle_on_td_change(self, req):
        """Handler for the 'on_td_change' subscription method."""

        params = req.params

        try:
            validate(params, SCHEMA_PARAMS_ON_TD_CHANGE)
        except ValidationError as ex:
            self._write_error(str(ex), WebsocketErrors.INVALID_METHOD_PARAMS, msg_id=req.id)
            return

        subscription_id = str(uuid.uuid4())

        res = WebsocketMessageResponse(result=subscription_id, msg_id=req.id)
        self.write_message(res.to_json())

        observable = self.exposed_thing.on_td_change()

        self._subscribe(subscription_id, observable)

    @gen.coroutine
    def _handle_on_event(self, req):
        """Handler for the 'on_event' subscription method."""

        params = req.params

        try:
            validate(params, SCHEMA_PARAMS_ON_EVENT)
        except ValidationError as ex:
            self._write_error(str(ex), WebsocketErrors.INVALID_METHOD_PARAMS, msg_id=req.id)
            return

        subscription_id = str(uuid.uuid4())

        res = WebsocketMessageResponse(result=subscription_id, msg_id=req.id)
        self.write_message(res.to_json())

        observable = self.exposed_thing.on_event(name=params["name"])

        self._subscribe(subscription_id, observable)

    @gen.coroutine
    def _handle_dispose(self, req):
        """Handler for the 'dispose' method."""

        params = req.params

        try:
            validate(params, SCHEMA_PARAMS_DISPOSE)
        except ValidationError as ex:
            self._write_error(str(ex), WebsocketErrors.INVALID_METHOD_PARAMS, msg_id=req.id)
            return

        result = None
        subscription_id = params["subscription"]

        if subscription_id in self._subscriptions:
            self._dispose_subscription(subscription_id)
            result = subscription_id

        res = WebsocketMessageResponse(result=result, msg_id=req.id)
        self.write_message(res.to_json())

    @gen.coroutine
    def _handle(self, req):
        """Takes a WebsocketMessageRequest instance and routes
        the request to the required method handler."""

        handler_map = {
            WebsocketMethods.READ_PROPERTY: self._handle_get_property,
            WebsocketMethods.WRITE_PROPERTY: self._handle_set_property,
            WebsocketMethods.INVOKE_ACTION: self._handle_invoke_action,
            WebsocketMethods.ON_PROPERTY_CHANGE: self._handle_on_property_change,
            WebsocketMethods.ON_TD_CHANGE: self._handle_on_td_change,
            WebsocketMethods.ON_EVENT: self._handle_on_event,
            WebsocketMethods.DISPOSE: self._handle_dispose,
        }

        if req.method not in handler_map:
            self._write_error("Unimplemented method", WebsocketErrors.INTERNAL_ERROR, msg_id=req.id)
            return

        handler = handler_map[req.method]
        yield handler(req)

    @gen.coroutine
    def on_message(self, message):
        """Called each time the server receives a WebSockets message.
        All messages that do not conform to the protocol are discarded."""

        try:
            req = WebsocketMessageRequest.from_raw(message)
            gen.convert_yielded(self._handle(req))
        except WebsocketMessageException as ex:
            self._write_error(str(ex), WebsocketErrors.INTERNAL_ERROR)

    def on_close(self):
        """Called when the WebSockets connection is closed."""

        for subscription_id in list(self._subscriptions.keys()):
            self._dispose_subscription(subscription_id)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Class that implements the WebSockets server.
"""

import tornado.gen
from tornado import web
from tornado.httpserver import HTTPServer

from wotpy.codecs.enums import MediaTypes
from wotpy.protocols.enums import Protocols
from wotpy.protocols.server import BaseProtocolServer
from wotpy.protocols.ws.enums import WebsocketSchemes
from wotpy.protocols.ws.handler import WebsocketHandler
from wotpy.wot.form import Form


class WebsocketServer(BaseProtocolServer):
    """WebSockets binding server implementation. Builds a Tornado application
    that uses the WebsocketHandler handler to process WebSockets messages."""

    DEFAULT_PORT = 81

    def __init__(self, port=DEFAULT_PORT, ssl_context=None):
        super(WebsocketServer, self).__init__(port=port)
        self._server = None
        self._app = self._build_app()
        self._ssl_context = ssl_context

    @property
    def protocol(self):
        """Protocol of this server instance.
        A member of the Protocols enum."""

        return Protocols.WEBSOCKETS

    @property
    def scheme(self):
        """Returns the URL scheme for this server."""

        return WebsocketSchemes.WSS if self.is_secure else WebsocketSchemes.WS

    @property
    def is_secure(self):
        """Returns True if this server is configured to use SSL encryption."""

        return self._ssl_context is not None

    @property
    def app(self):
        """Tornado application property."""

        return self._app

    def _build_app(self):
        """Builds and returns the Tornado application for the WebSockets server."""

        return web.Application([(
            r"/(?P<name>[^\/]+)",
            WebsocketHandler,
            {"websocket_server": self}
        )])

    def build_forms(self, hostname, interaction):
        """Builds and returns a list with all Form that are
        linked to this server for the given Interaction."""

        exposed_thing = self.exposed_thing_set.find_by_interaction(interaction)

        if not exposed_thing:
            raise ValueError("Unknown Interaction")

        base_url = self.build_base_url(hostname=hostname, thing=exposed_thing.thing)

        return [
            Form(
                interaction=interaction,
                protocol=self.protocol,
                href=base_url,
                content_type=MediaTypes.JSON)
        ]

    def build_base_url(self, hostname, thing):
        """Returns the base URL for the given Thing in the context of this server."""

        if not self.exposed_thing_set.find_by_thing_id(thing.id):
            raise ValueError("Unknown Thing")

        hostname = hostname.rstrip("/")

        return "{}://{}:{}/{}".format(self.scheme, hostname, self.port, thing.url_name)

    @tornado.gen.coroutine
    def start(self):
        """Starts the WebSockets server."""

        self._server = HTTPServer(self.app, ssl_options=self._ssl_context)
        self._server.listen(self.port)

    @tornado.gen.coroutine
    def stop(self):
        """Stops the WebSockets server."""

        if not self._server:
            return

        self._server.stop()
        self._server = None



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Classes that represent Interaction instances accessed on a ConsumedThing.
"""

import six
import tornado.gen
from rx.concurrency import IOLoopScheduler
from six.moves import UserDict
from slugify import slugify


class ConsumedThingInteractionDict(UserDict):
    """A dictionary that provides lazy access to the objects that implement
    the Interaction interface for each interaction in a given ConsumedThing."""

    def __init__(self, *args, **kwargs):
        self._consumed_thing = kwargs.pop("consumed_thing")
        UserDict.__init__(self, *args, **kwargs)

    def _find_normalized_name(self, name):
        """Takes a case-insensitive URL-safe interaction name and returns
        the actual name in the interaction dict."""

        return next((key for key in six.iterkeys(self.interaction_dict) if slugify(key) == slugify(name)), None)

    def __getitem__(self, name):
        """Lazily build and return an object that implements the Interaction interface."""

        name_normalized = self._find_normalized_name(name)

        if name_normalized is None:
            raise KeyError("Unknown interaction: {}".format(name))

        return self.thing_interaction_class(self._consumed_thing, name_normalized)

    def __len__(self):
        return len(self.interaction_dict)

    def __contains__(self, item):
        return self._find_normalized_name(item) is not None

    def __iter__(self):
        return six.iterkeys(self.interaction_dict)

    @property
    def interaction_dict(self):
        """Returns an interactions dict by name.
        The dict values are the raw dict interactions as contained in a TD document."""

        raise NotImplementedError()

    @property
    def thing_interaction_class(self):
        """Returns the class that implements the
        Interaction interface for this type of interaction."""

        raise NotImplementedError()


class ConsumedThingPropertyDict(ConsumedThingInteractionDict):
    """A dictionary that provides lazy access to the objects that implement
    the ThingProperty interface for each property in a given ConsumedThing."""

    @property
    def interaction_dict(self):
        return self._consumed_thing.td.properties

    @property
    def thing_interaction_class(self):
        return ConsumedThingProperty


class ConsumedThingActionDict(ConsumedThingInteractionDict):
    """A dictionary that provides lazy access to the objects that implement
    the ThingAction interface for each action in a given ConsumedThing."""

    @property
    def interaction_dict(self):
        return self._consumed_thing.td.actions

    @property
    def thing_interaction_class(self):
        return ConsumedThingAction


class ConsumedThingEventDict(ConsumedThingInteractionDict):
    """A dictionary that provides lazy access to the objects that implement
    the ThingEvent interface for each event in a given ConsumedThing."""

    @property
    def interaction_dict(self):
        return self._consumed_thing.td.events

    @property
    def thing_interaction_class(self):
        return ConsumedThingEvent


class ConsumedThingProperty(object):
    """The ThingProperty interface implementation for ConsumedThing objects."""

    def __init__(self, consumed_thing, name):
        self._consumed_thing = consumed_thing
        self._name = name

    def __str__(self):
        return "<{}> ({}::{})".format(self.__class__.__name__, self._consumed_thing.id, self._name)

    def __getattr__(self, name):
        """Search for members that raised an AttributeError in
        the private init dict before propagating the exception."""

        return getattr(self._consumed_thing.td.properties[self._name], name)

    @tornado.gen.coroutine
    def read(self, timeout=None, client_kwargs=None):
        """The read() method will fetch the value of the Property.
        A coroutine that yields the value or raises an error."""

        value = yield self._consumed_thing.read_property(
            self._name,
            timeout=timeout,
            client_kwargs=client_kwargs)

        raise tornado.gen.Return(value)

    @tornado.gen.coroutine
    def write(self, value, timeout=None, client_kwargs=None):
        """The write() method will attempt to set the value of the
        Property specified in the value argument whose type SHOULD
        match the one specified by the type property.
        A coroutine that yields on success or raises an error."""

        yield self._consumed_thing.write_property(
            self._name, value,
            timeout=timeout,
            client_kwargs=client_kwargs)

    def subscribe(self, *args, **kwargs):
        """Subscribe to an stream of events emitted when the property value changes."""

        client_kwargs = kwargs.pop("client_kwargs", None)

        observable = self._consumed_thing.on_property_change(
            self._name,
            client_kwargs=client_kwargs)

        return observable.subscribe_on(IOLoopScheduler()).subscribe(*args, **kwargs)


class ConsumedThingAction(object):
    """The ThingAction interface implementation for ConsumedThing objects."""

    def __init__(self, consumed_thing, name):
        self._consumed_thing = consumed_thing
        self._name = name

    def __str__(self):
        return "<{}> ({}::{})".format(self.__class__.__name__, self._consumed_thing.id, self._name)

    def __getattr__(self, name):
        """Search for members that raised an AttributeError in
        the private init dict before propagating the exception."""

        return getattr(self._consumed_thing.td.actions[self._name], name)

    @tornado.gen.coroutine
    def invoke(self, *args, **kwargs):
        """The invoke() method when invoked, starts the Action interaction
        with the input value provided by the inputValue argument."""

        input_value = args[0] if len(args) else None
        client_kwargs = kwargs.pop("client_kwargs", None)
        timeout = kwargs.pop("timeout", None)

        result = yield self._consumed_thing.invoke_action(
            self._name, input_value,
            timeout=timeout,
            client_kwargs=client_kwargs)

        raise tornado.gen.Return(result)


class ConsumedThingEvent(object):
    """The ThingEvent interface implementation for ConsumedThing objects."""

    def __init__(self, consumed_thing, name):
        self._consumed_thing = consumed_thing
        self._name = name

    def __str__(self):
        return "<{}> ({}::{})".format(self.__class__.__name__, self._consumed_thing.id, self._name)

    def __getattr__(self, name):
        """Search for members that raised an AttributeError in
        the private init dict before propagating the exception."""

        return getattr(self._consumed_thing.td.events[self._name], name)

    def subscribe(self, *args, **kwargs):
        """Subscribe to an stream of emissions of this event."""

        client_kwargs = kwargs.pop("client_kwargs", None)

        observable = self._consumed_thing.on_event(
            self._name,
            client_kwargs=client_kwargs)

        return observable.subscribe_on(IOLoopScheduler()).subscribe(*args, **kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import uuid

import pytest
import tornado.gen
import tornado.ioloop
from faker import Faker

from tests.utils import find_free_port
from wotpy.protocols.http.server import HTTPServer
from wotpy.wot.dictionaries.interaction import PropertyFragmentDict, ActionFragmentDict, EventFragmentDict
from wotpy.wot.exposed.thing import ExposedThing
from wotpy.wot.servient import Servient
from wotpy.wot.td import ThingDescription
from wotpy.wot.thing import Thing


@pytest.fixture
def http_server():
    """Builds an HTTPServer instance that contains an ExposedThing."""

    exposed_thing = ExposedThing(servient=Servient(), thing=Thing(id=uuid.uuid4().urn))

    exposed_thing.add_property(uuid.uuid4().hex, PropertyFragmentDict({
        "type": "number",
        "observable": True
    }), value=Faker().pyint())

    exposed_thing.add_property(uuid.uuid4().hex, PropertyFragmentDict({
        "type": "number",
        "observable": True
    }), value=Faker().pyint())

    exposed_thing.add_event(uuid.uuid4().hex, EventFragmentDict({
        "type": "object"
    }))

    action_name = uuid.uuid4().hex

    @tornado.gen.coroutine
    def triple(parameters):
        input_value = parameters.get("input")
        yield tornado.gen.sleep(0)
        raise tornado.gen.Return(input_value * 3)

    exposed_thing.add_action(action_name, ActionFragmentDict({
        "input": {"type": "number"},
        "output": {"type": "number"}
    }), triple)

    port = find_free_port()

    server = HTTPServer(port=port)
    server.add_exposed_thing(exposed_thing)

    @tornado.gen.coroutine
    def start():
        yield server.start()

    tornado.ioloop.IOLoop.current().run_sync(start)

    yield server

    @tornado.gen.coroutine
    def stop():
        yield server.stop()

    tornado.ioloop.IOLoop.current().run_sync(stop)


@pytest.fixture
def http_servient():
    """Returns a Servient that exposes an HTTP server and one ExposedThing."""

    http_port = find_free_port()
    http_server = HTTPServer(port=http_port)

    servient = Servient(catalogue_port=None)
    servient.add_server(http_server)

    @tornado.gen.coroutine
    def start():
        raise tornado.gen.Return((yield servient.start()))

    wot = tornado.ioloop.IOLoop.current().run_sync(start)

    property_name_01 = uuid.uuid4().hex
    property_name_02 = uuid.uuid4().hex
    action_name_01 = uuid.uuid4().hex
    event_name_01 = uuid.uuid4().hex

    td_dict = {
        "id": uuid.uuid4().urn,
        "name": uuid.uuid4().hex,
        "properties": {
            property_name_01: {
                "observable": True,
                "type": "string"
            },
            property_name_02: {
                "observable": True,
                "type": "string"
            }
        },
        "actions": {
            action_name_01: {
                "input": {
                    "type": "number"
                },
                "output": {
                    "type": "number"
                },
            }
        },
        "events": {
            event_name_01: {
                "type": "string"
            }
        },
    }

    td = ThingDescription(td_dict)

    exposed_thing = wot.produce(td.to_str())
    exposed_thing.expose()

    @tornado.gen.coroutine
    def action_handler(parameters):
        input_value = parameters.get("input")
        raise tornado.gen.Return(int(input_value) * 2)

    exposed_thing.set_action_handler(action_name_01, action_handler)

    yield servient

    @tornado.gen.coroutine
    def shutdown():
        yield servient.shutdown()

    tornado.ioloop.IOLoop.current().run_sync(shutdown)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
MQTT Protocol Binding implementation.

.. autosummary::
    :toctree: _mqtt

    wotpy.protocols.mqtt.handlers
    wotpy.protocols.mqtt.client
    wotpy.protocols.mqtt.enums
    wotpy.protocols.mqtt.runner
    wotpy.protocols.mqtt.server
"""

from wotpy.support import is_mqtt_supported

if is_mqtt_supported() is False:
    raise NotImplementedError("MQTT binding is not supported in this platform")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import ssl
import tempfile
import uuid

import pytest
import tornado.gen
import tornado.ioloop
from OpenSSL import crypto
from faker import Faker

from tests.utils import find_free_port
from wotpy.protocols.http.server import HTTPServer
from wotpy.protocols.ws.server import WebsocketServer
from wotpy.support import is_coap_supported, is_mqtt_supported
from wotpy.wot.servient import Servient
from wotpy.wot.td import ThingDescription


@pytest.fixture
def all_protocols_servient():
    """Returns a Servient configured to use all available protocol bindings."""

    servient = Servient(catalogue_port=None)

    http_port = find_free_port()
    http_server = HTTPServer(port=http_port)
    servient.add_server(http_server)

    ws_port = find_free_port()
    ws_server = WebsocketServer(port=ws_port)
    servient.add_server(ws_server)

    if is_coap_supported():
        from wotpy.protocols.coap.server import CoAPServer
        coap_port = find_free_port()
        coap_server = CoAPServer(port=coap_port)
        servient.add_server(coap_server)

    if is_mqtt_supported():
        from wotpy.protocols.mqtt.server import MQTTServer
        from tests.protocols.mqtt.broker import get_test_broker_url, is_test_broker_online
        if is_test_broker_online():
            mqtt_server = MQTTServer(broker_url=get_test_broker_url())
            servient.add_server(mqtt_server)

    @tornado.gen.coroutine
    def start():
        raise tornado.gen.Return((yield servient.start()))

    wot = tornado.ioloop.IOLoop.current().run_sync(start)

    td_dict = {
        "id": uuid.uuid4().urn,
        "name": uuid.uuid4().hex,
        "properties": {
            uuid.uuid4().hex: {
                "observable": True,
                "type": "string"
            }
        }
    }

    td = ThingDescription(td_dict)

    exposed_thing = wot.produce(td.to_str())
    exposed_thing.expose()

    yield servient

    @tornado.gen.coroutine
    def shutdown():
        yield servient.shutdown()

    tornado.ioloop.IOLoop.current().run_sync(shutdown)


@pytest.fixture
def self_signed_ssl_context():
    """Returns a self-signed SSL certificate."""

    base_dir = tempfile.gettempdir()

    certfile = os.path.join(base_dir, "{}.pem".format(uuid.uuid4().hex))
    keyfile = os.path.join(base_dir, "{}.pem".format(uuid.uuid4().hex))

    pkey = crypto.PKey()
    pkey.generate_key(crypto.TYPE_RSA, 2048)

    cert = crypto.X509()
    cert.get_subject().C = "ES"
    cert.get_subject().ST = Faker().pystr()
    cert.get_subject().L = Faker().pystr()
    cert.get_subject().O = Faker().pystr()
    cert.get_subject().OU = Faker().pystr()
    cert.get_subject().CN = Faker().pystr()
    cert.set_serial_number(Faker().pyint())
    cert.gmtime_adj_notBefore(0)
    cert.gmtime_adj_notAfter(3600)
    cert.set_issuer(cert.get_subject())
    cert.set_pubkey(pkey)
    # noinspection PyTypeChecker
    cert.sign(pkey, "sha384")

    with open(certfile, "wb") as fh:
        fh.write(crypto.dump_certificate(crypto.FILETYPE_PEM, cert))

    with open(keyfile, "wb") as fh:
        fh.write(crypto.dump_privatekey(crypto.FILETYPE_PEM, pkey))

    ssl_context = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
    ssl_context.load_cert_chain(certfile=certfile, keyfile=keyfile)

    yield ssl_context

    os.remove(certfile)
    os.remove(keyfile)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Classes that represent JSON-RPC messages exchanged over WebSockets.
"""

import json

from jsonschema import validate, ValidationError

from wotpy.protocols.ws.enums import WebsocketErrors
from wotpy.protocols.ws.schemas import \
    SCHEMA_REQUEST, \
    SCHEMA_RESPONSE, \
    SCHEMA_EMITTED_ITEM, \
    SCHEMA_ERROR, \
    JSON_RPC_VERSION
from wotpy.utils.utils import to_json_obj


def parse_ws_message(raw_msg):
    """Takes a raw WebSockets message and attempts
    to parse it to create a message instance."""

    msg_klasses = [
        WebsocketMessageRequest,
        WebsocketMessageError,
        WebsocketMessageResponse,
        WebsocketMessageEmittedItem
    ]

    for klass in msg_klasses:
        try:
            msg_instance = klass.from_raw(raw_msg)
            return msg_instance
        except WebsocketMessageException:
            pass

    raise WebsocketMessageException("Invalid message: {}".format(raw_msg))


class WebsocketMessageException(Exception):
    """Exception raised when a WS message appears to be invalid."""

    pass


class WebsocketMessageRequest(object):
    """Represents a message received on a websocket that
    contains a JSON-RPC WoT action request."""

    @classmethod
    def from_raw(cls, raw_msg):
        """Builds a new WebsocketMessageRequest instance from a raw socket message.
        Raises WebsocketMessageException if the message is invalid."""

        try:
            msg = json.loads(raw_msg)
            validate(msg, SCHEMA_REQUEST)

            return WebsocketMessageRequest(
                method=msg["method"],
                params=msg["params"],
                msg_id=msg.get("id", None))
        except Exception as ex:
            raise WebsocketMessageException(str(ex))

    def __init__(self, method, params, msg_id=None):
        self.method = method
        self.params = params
        self.msg_id = msg_id

        try:
            validate(self.to_dict(), SCHEMA_REQUEST)
        except ValidationError as ex:
            raise WebsocketMessageError(str(ex))

    @property
    def id(self):
        """ID property."""

        return self.msg_id

    def to_dict(self):
        """Returns this message as a dict."""

        msg = {
            "jsonrpc": JSON_RPC_VERSION,
            "method": self.method,
            "params": self.params,
            "id": self.id
        }

        return msg

    def to_json(self):
        """Returns this message as a JSON string."""

        return json.dumps(self.to_dict())


class WebsocketMessageResponse(object):
    """Represents a WoT Websockets JSON-RPC response message."""

    @classmethod
    def from_raw(cls, raw_msg):
        """Builds a new WebsocketMessageResponse instance from a raw socket message.
        Raises WebsocketMessageException if the message is invalid."""

        try:
            msg = json.loads(raw_msg)
            validate(msg, SCHEMA_RESPONSE)

            return WebsocketMessageResponse(
                result=msg["result"],
                msg_id=msg.get("id", None))
        except Exception as ex:
            raise WebsocketMessageException(str(ex))

    def __init__(self, result, msg_id=None):
        self.result = result
        self.msg_id = msg_id

        try:
            validate(self.to_dict(), SCHEMA_RESPONSE)
        except ValidationError as ex:
            raise WebsocketMessageError(str(ex))

    @property
    def id(self):
        """ID property."""

        return self.msg_id

    def to_dict(self):
        """Returns this message as a dict."""

        msg = {
            "jsonrpc": JSON_RPC_VERSION,
            "result": self.result,
            "id": self.id
        }

        return msg

    def to_json(self):
        """Returns this message as a JSON string."""

        return json.dumps(self.to_dict())


class WebsocketMessageError(object):
    """Represents a WoT Websockets JSON-RPC error message."""

    @classmethod
    def from_raw(cls, raw_msg):
        """Builds a new WebsocketMessageError instance from a raw socket message.
        Raises WebsocketMessageException if the message is invalid."""

        try:
            msg = json.loads(raw_msg)
            validate(msg, SCHEMA_ERROR)

            return WebsocketMessageError(
                message=msg["error"]["message"],
                code=msg["error"]["code"],
                data=msg["error"].get("data", None),
                msg_id=msg.get("id", None))
        except Exception as ex:
            raise WebsocketMessageException(str(ex))

    def __init__(self, message, code=WebsocketErrors.INTERNAL_ERROR, data=None, msg_id=None):
        self.message = message
        self.msg_id = msg_id
        self.code = code
        self.data = data

        try:
            validate(self.to_dict(), SCHEMA_ERROR)
        except ValidationError as ex:
            raise WebsocketMessageError(str(ex))

    @property
    def id(self):
        """ID property."""

        return self.msg_id

    def to_dict(self):
        """Returns this message as a dict."""

        msg = {
            "jsonrpc": JSON_RPC_VERSION,
            "error": {
                "code": self.code,
                "message": self.message,
                "data": self.data
            },
            "id": self.id
        }

        return msg

    def to_json(self):
        """Returns this message as a JSON string."""

        return json.dumps(self.to_dict())


class WebsocketMessageEmittedItem(object):
    """Represents a Websockets message for an item emitted by an active subscription."""

    @classmethod
    def from_raw(cls, raw_msg):
        """Builds a new WebsocketMessageEmittedItem instance from a raw socket message.
        Raises WebsocketMessageException if the message is invalid."""

        try:
            msg = json.loads(raw_msg)
            validate(msg, SCHEMA_EMITTED_ITEM)

            return WebsocketMessageEmittedItem(
                subscription_id=msg["subscription"],
                name=msg["name"],
                data=msg["data"])
        except Exception as ex:
            raise WebsocketMessageException(str(ex))

    def __init__(self, subscription_id, name, data):
        self.subscription_id = subscription_id
        self.name = name
        self.data = to_json_obj(data)

        try:
            validate(self.to_dict(), SCHEMA_EMITTED_ITEM)
        except ValidationError as ex:
            raise WebsocketMessageError(str(ex))

    def to_dict(self):
        """Returns this message as a dict."""

        msg = {
            "subscription": self.subscription_id,
            "name": self.name,
            "data": self.data
        }

        return msg

    def to_json(self):
        """Returns this message as a JSON string."""

        return json.dumps(self.to_dict())



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import random
import uuid

import pytest
import six
import tornado.concurrent
import tornado.gen
import tornado.ioloop
import tornado.locks
import tornado.websocket
from mock import patch
from rx.concurrency import IOLoopScheduler
from tornado.concurrent import Future

from tests.protocols.helpers import \
    client_test_on_event, \
    client_test_read_property, \
    client_test_write_property, \
    client_test_invoke_action, \
    client_test_invoke_action_error, \
    client_test_on_property_change_error
from tests.utils import run_test_coroutine
from wotpy.protocols.exceptions import ProtocolClientException, ClientRequestTimeout
from wotpy.protocols.ws.client import WebsocketClient
from wotpy.wot.td import ThingDescription


def test_read_property(websocket_servient):
    """The Websockets client can read properties."""

    client_test_read_property(websocket_servient, WebsocketClient)


def test_read_property_unknown(websocket_servient):
    """The Websockets client raises an error when attempting to read an unknown property."""

    exposed_thing = next(websocket_servient.exposed_things)
    td = ThingDescription.from_thing(exposed_thing.thing)

    @tornado.gen.coroutine
    def test_coroutine():
        ws_client = WebsocketClient()

        with pytest.raises(ProtocolClientException):
            yield ws_client.read_property(td, uuid.uuid4().hex)

    run_test_coroutine(test_coroutine)


def test_write_property(websocket_servient):
    """The Websockets client can write properties."""

    client_test_write_property(websocket_servient, WebsocketClient)


def test_invoke_action(websocket_servient):
    """The Websockets client can invoke actions."""

    client_test_invoke_action(websocket_servient, WebsocketClient)


def test_invoke_action_error(websocket_servient):
    """Errors raised by Actions are propagated propertly by the WebSockets binding client."""

    client_test_invoke_action_error(websocket_servient, WebsocketClient)


def test_on_event(websocket_servient):
    """The Websockets client can observe events."""

    client_test_on_event(websocket_servient, WebsocketClient)


def test_on_property_change(websocket_servient):
    """The Websockets client can observe property changes."""

    exposed_thing = next(websocket_servient.exposed_things)
    td = ThingDescription.from_thing(exposed_thing.thing)

    @tornado.gen.coroutine
    def test_coroutine():
        ws_client = WebsocketClient()

        prop_names = list(td.properties.keys())
        prop_name_01 = prop_names[0]
        prop_name_02 = prop_names[1]

        obsv_01 = ws_client.on_property_change(td, prop_name_01)
        obsv_02 = ws_client.on_property_change(td, prop_name_02)

        prop_values_01 = [uuid.uuid4().hex for _ in range(10)]
        prop_values_02 = [uuid.uuid4().hex for _ in range(90)]

        future_values_01 = {key: Future() for key in prop_values_01}
        future_values_02 = {key: Future() for key in prop_values_02}

        future_conn_01 = Future()
        future_conn_02 = Future()

        def build_on_next(fut_conn, fut_vals):
            def on_next(ev):
                if not fut_conn.done():
                    fut_conn.set_result(True)

                if ev.data.value in fut_vals:
                    fut_vals[ev.data.value].set_result(True)

            return on_next

        on_next_01 = build_on_next(future_conn_01, future_values_01)
        on_next_02 = build_on_next(future_conn_02, future_values_02)

        subscription_01 = obsv_01.subscribe_on(IOLoopScheduler()).subscribe(on_next_01)
        subscription_02 = obsv_02.subscribe_on(IOLoopScheduler()).subscribe(on_next_02)

        while not future_conn_01.done() or not future_conn_02.done():
            yield exposed_thing.write_property(prop_name_01, uuid.uuid4().hex)
            yield exposed_thing.write_property(prop_name_02, uuid.uuid4().hex)
            yield tornado.gen.sleep(0)

        assert len(prop_values_01) < len(prop_values_02)

        for idx in range(len(prop_values_01)):
            yield exposed_thing.write_property(prop_name_01, prop_values_01[idx])
            yield exposed_thing.write_property(prop_name_02, prop_values_02[idx])

        yield list(future_values_01.values())

        assert next(fut for fut in six.itervalues(future_values_02) if not fut.done())

        subscription_01.dispose()

        for val in prop_values_02[len(prop_values_01):]:
            yield exposed_thing.write_property(prop_name_02, val)

        yield list(future_values_02.values())

        subscription_02.dispose()

    run_test_coroutine(test_coroutine)


def test_on_property_change_error(websocket_servient):
    """Errors that arise in the middle of an ongoing Property
    observation are propagated to the subscription as expected."""

    client_test_on_property_change_error(websocket_servient, WebsocketClient)


# noinspection PyUnusedLocal
def _condition_coro(*args, **kwargs):
    """Coroutine mock side effect that returns a Condition that is never notified."""

    @tornado.gen.coroutine
    def _coro():
        raise tornado.gen.Return(tornado.locks.Condition())

    return _coro()


def test_timeout_read_property(websocket_servient):
    """Timeouts can be defined on Property reads."""

    # noinspection PyUnresolvedReferences
    with patch.object(WebsocketClient, '_send_message', _condition_coro):
        with pytest.raises(ClientRequestTimeout):
            client_test_read_property(websocket_servient, WebsocketClient, timeout=random.random())


def test_timeout_write_property(websocket_servient):
    """Timeouts can be defined on Property writes."""

    # noinspection PyUnresolvedReferences
    with patch.object(WebsocketClient, '_send_message', _condition_coro):
        with pytest.raises(ClientRequestTimeout):
            client_test_write_property(websocket_servient, WebsocketClient, timeout=random.random())


def test_timeout_invoke_action(websocket_servient):
    """Timeouts can be defined on Action invocations."""

    # noinspection PyUnresolvedReferences
    with patch.object(WebsocketClient, '_send_message', _condition_coro):
        with pytest.raises(ClientRequestTimeout):
            client_test_invoke_action(websocket_servient, WebsocketClient, timeout=random.random())



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from tests.protocols.helpers import \
    client_test_on_property_change, \
    client_test_on_event, \
    client_test_read_property, \
    client_test_write_property, \
    client_test_invoke_action, \
    client_test_invoke_action_error, \
    client_test_on_property_change_error
from wotpy.protocols.http.client import HTTPClient


def test_read_property(http_servient):
    """The HTTP client can read properties."""

    client_test_read_property(http_servient, HTTPClient)


def test_write_property(http_servient):
    """The HTTP client can write properties."""

    client_test_write_property(http_servient, HTTPClient)


def test_invoke_action(http_servient):
    """The HTTP client can invoke actions."""

    client_test_invoke_action(http_servient, HTTPClient)


def test_invoke_action_error(http_servient):
    """Errors raised by Actions are propagated propertly by the HTTP binding client."""

    client_test_invoke_action_error(http_servient, HTTPClient)


def test_on_event(http_servient):
    """The HTTP client can subscribe to event emissions."""

    client_test_on_event(http_servient, HTTPClient)


def test_on_property_change(http_servient):
    """The HTTP client can subscribe to property updates."""

    client_test_on_property_change(http_servient, HTTPClient)


def test_on_property_change_error(http_servient):
    """Errors that arise in the middle of an ongoing Property
    observation are propagated to the subscription as expected."""

    client_test_on_property_change_error(http_servient, HTTPClient)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Request handler for Action interactions.
"""

import logging
import pprint
import time
import uuid

import six
import tornado.gen
from tornado.web import HTTPError
from tornado.web import RequestHandler

import wotpy.protocols.http.handlers.utils as handler_utils


# noinspection PyAbstractClass,PyAttributeOutsideInit
class ActionInvokeHandler(RequestHandler):
    """Handler for Action invocation requests."""

    # noinspection PyMethodOverriding
    def initialize(self, http_server):
        self._server = http_server

    @tornado.gen.coroutine
    def post(self, thing_name, name):
        """Invokes the action and returns the invocation result."""

        exposed_thing = handler_utils.get_exposed_thing(self._server, thing_name)
        input_value = handler_utils.get_argument(self, "input")
        future_result = exposed_thing.actions[name].invoke(input_value)
        invocation_id = uuid.uuid4().hex
        self._server.pending_actions[invocation_id] = future_result
        self.write({"invocation": "/invocation/{}".format(invocation_id)})


# noinspection PyAbstractClass,PyAttributeOutsideInit
class PendingInvocationHandler(RequestHandler):
    """Handler to check the status of pending action invocations."""

    # noinspection PyMethodOverriding
    def initialize(self, http_server):
        self._server = http_server
        self._logr = logging.getLogger(__name__)

    def _clean_expired(self):
        """Removes the Action invocations that are expired and
        have already been checked by at least one client."""

        now = time.time()

        expired_invocations = [
            inv_id for inv_id, tstamp in six.iteritems(self._server.invocation_check_times)
            if (now - tstamp) > self._server.action_ttl
        ]

        if len(expired_invocations):
            self._logr.debug("Expired invocations: {}".format(pprint.pformat(expired_invocations)))

        for invocation_id in expired_invocations:
            self._server.invocation_check_times.pop(invocation_id)
            fut_result = self._server.pending_actions.get(invocation_id, None)

            if fut_result and fut_result.done():
                self._logr.debug("Removing completed invocation Future: {}".format(invocation_id))
                self._server.pending_actions.pop(invocation_id, None)

    @tornado.gen.coroutine
    def get(self, invocation_id):
        """Checks and returns the status of the Future that represents an action invocation."""

        if invocation_id not in self._server.pending_actions:
            raise HTTPError(log_message="Unknown invocation: {}".format(invocation_id))

        try:
            result = yield self._server.pending_actions[invocation_id]
            self.write({"done": True, "result": result})
        except Exception as ex:
            self.write({"done": True, "error": str(ex)})
        finally:
            self._logr.debug("Updating invocation check time: {}".format(invocation_id))
            self._server.invocation_check_times[invocation_id] = time.time()
            self._clean_expired()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import asyncio
import random

import pytest
import six
import tornado.gen
from faker import Faker
from mock import MagicMock, patch

from tests.protocols.helpers import \
    client_test_on_property_change, \
    client_test_on_event, \
    client_test_read_property, \
    client_test_write_property, \
    client_test_invoke_action, \
    client_test_invoke_action_error
from tests.protocols.mqtt.broker import is_test_broker_online, BROKER_SKIP_REASON
from tests.utils import run_test_coroutine, DEFAULT_TIMEOUT_SECS
from wotpy.protocols.exceptions import ClientRequestTimeout
from wotpy.protocols.mqtt.client import MQTTClient
from wotpy.wot.td import ThingDescription

pytestmark = pytest.mark.skipif(is_test_broker_online() is False, reason=BROKER_SKIP_REASON)


def test_read_property(mqtt_servient):
    """Property values may be retrieved using the MQTT binding client."""

    client_test_read_property(mqtt_servient, MQTTClient)


def test_write_property(mqtt_servient):
    """Properties may be updated using the MQTT binding client."""

    client_test_write_property(mqtt_servient, MQTTClient)


def test_invoke_action(mqtt_servient):
    """Actions may be invoked using the MQTT binding client."""

    client_test_invoke_action(mqtt_servient, MQTTClient)


def test_invoke_action_error(mqtt_servient):
    """Errors raised by Actions are propagated propertly by the MQTT binding client."""

    client_test_invoke_action_error(mqtt_servient, MQTTClient)


def test_on_property_change(mqtt_servient):
    """Property updates may be observed using the MQTT binding client."""

    client_test_on_property_change(mqtt_servient, MQTTClient)


def test_on_event(mqtt_servient):
    """Event emissions may be observed using the MQTT binding client."""

    client_test_on_event(mqtt_servient, MQTTClient)


# noinspection PyUnusedLocal
def _effect_dummy(*args, **kwargs):
    """Coroutine mock side effect that does nothing and returns a Mock."""

    @tornado.gen.coroutine
    def _coro():
        yield tornado.gen.moment
        raise tornado.gen.Return(MagicMock())

    return _coro()


# noinspection PyUnusedLocal
def _effect_raise_timeout(*args, **kwargs):
    """Coroutine mock side effect that raises a timeout error."""

    @tornado.gen.coroutine
    def _coro():
        yield tornado.gen.moment
        raise asyncio.TimeoutError

    return _coro()


def _build_effect_sleep(sleep_secs):
    """Factory function to build coroutine mock side effects to sleep a fixed amount of time."""

    # noinspection PyUnusedLocal
    def _effect_wait(*args, **kwargs):
        @tornado.gen.coroutine
        def _coro():
            yield tornado.gen.sleep(sleep_secs)
            raise asyncio.TimeoutError

        return _coro()

    return _effect_wait


def _build_hbmqtt_mock(side_effect_deliver_message):
    """Returns a mock of the HBMQTT Client class."""

    mock_client = MagicMock()
    mock_client.connect.side_effect = _effect_dummy
    mock_client.deliver_message.side_effect = side_effect_deliver_message
    mock_client.disconnect.side_effect = _effect_dummy
    mock_client.subscribe.side_effect = _effect_dummy
    mock_client.publish.side_effect = _effect_dummy

    mock_cls = MagicMock()
    mock_cls.return_value = mock_client

    return mock_cls


def test_timeout_invoke_action(mqtt_servient):
    """Timeouts can be defined on Action invocations."""

    exposed_thing = next(mqtt_servient.exposed_things)
    action_name = next(six.iterkeys(exposed_thing.actions))
    td = ThingDescription.from_thing(exposed_thing.thing)
    mqtt_mock = _build_hbmqtt_mock(_effect_raise_timeout)

    timeout = random.random()

    @tornado.gen.coroutine
    def test_coroutine():
        with patch('wotpy.protocols.mqtt.client.hbmqtt.client.MQTTClient', new=mqtt_mock):
            mqtt_client = MQTTClient()

            with pytest.raises(ClientRequestTimeout):
                yield mqtt_client.invoke_action(td, action_name, Faker().pystr(), timeout=timeout)

    run_test_coroutine(test_coroutine)


def test_timeout_read_property(mqtt_servient):
    """Timeouts can be defined on Property reads."""

    exposed_thing = next(mqtt_servient.exposed_things)
    prop_name = next(six.iterkeys(exposed_thing.properties))
    td = ThingDescription.from_thing(exposed_thing.thing)
    mqtt_mock = _build_hbmqtt_mock(_effect_raise_timeout)

    timeout = random.random()

    @tornado.gen.coroutine
    def test_coroutine():
        with patch('wotpy.protocols.mqtt.client.hbmqtt.client.MQTTClient', new=mqtt_mock):
            mqtt_client = MQTTClient()

            with pytest.raises(ClientRequestTimeout):
                yield mqtt_client.read_property(td, prop_name, timeout=timeout)

    run_test_coroutine(test_coroutine)


def test_timeout_write_property(mqtt_servient):
    """Timeouts can be defined on Property writes."""

    exposed_thing = next(mqtt_servient.exposed_things)
    prop_name = next(six.iterkeys(exposed_thing.properties))
    td = ThingDescription.from_thing(exposed_thing.thing)
    mqtt_mock = _build_hbmqtt_mock(_effect_raise_timeout)

    timeout = random.random()

    @tornado.gen.coroutine
    def test_coroutine():
        with patch('wotpy.protocols.mqtt.client.hbmqtt.client.MQTTClient', new=mqtt_mock):
            mqtt_client = MQTTClient()

            with pytest.raises(ClientRequestTimeout):
                yield mqtt_client.write_property(td, prop_name, Faker().pystr(), timeout=timeout)

    run_test_coroutine(test_coroutine)


def test_stop_timeout(mqtt_servient):
    """Attempting to stop an unresponsive connection does not result in an indefinite wait."""

    exposed_thing = next(mqtt_servient.exposed_things)
    prop_name = next(six.iterkeys(exposed_thing.properties))
    td = ThingDescription.from_thing(exposed_thing.thing)

    timeout = random.random()

    assert (timeout * 3) < DEFAULT_TIMEOUT_SECS

    mqtt_mock = _build_hbmqtt_mock(_build_effect_sleep(DEFAULT_TIMEOUT_SECS * 10))

    @tornado.gen.coroutine
    def test_coroutine():
        with patch('wotpy.protocols.mqtt.client.hbmqtt.client.MQTTClient', new=mqtt_mock):
            mqtt_client = MQTTClient(stop_loop_timeout_secs=timeout)

            with pytest.raises(ClientRequestTimeout):
                yield mqtt_client.read_property(td, prop_name, timeout=timeout)

    run_test_coroutine(test_coroutine)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Base class for all MQTT handlers.
"""

import tornado.gen

from tornado.queues import Queue


class BaseMQTTHandler(object):
    """Base class for all MQTT handlers."""

    def __init__(self, mqtt_server):
        self._mqtt_server = mqtt_server
        self._queue = Queue()

    @property
    def servient_id(self):
        """Servient ID that is used to avoid topic collisions
        øwhen multiple Servients are connected to the same broker."""

        return self._mqtt_server.servient_id

    @property
    def mqtt_server(self):
        """MQTT server that contains this handler."""

        return self._mqtt_server

    @property
    def topics(self):
        """List of topics that this MQTT handler wants to subscribe to."""

        return None

    @property
    def queue(self):
        """Asynchronous queue where the handler leaves messages
        that should be published later by the runner."""

        return self._queue

    @tornado.gen.coroutine
    def handle_message(self, msg):
        """Called each time the runner receives a message for one of the handler topics."""

        pass

    @tornado.gen.coroutine
    def init(self):
        """Initializes the MQTT handler.
        Called when the MQTT runner starts."""

        pass

    @tornado.gen.coroutine
    def teardown(self):
        """Destroys the MQTT handler.
        Called when the MQTT runner stops."""

        pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Enumeration classes related to the various protocol servers.
"""

from wotpy.utils.enums import EnumListMixin


class Protocols(EnumListMixin):
    """Enumeration of protocol types."""

    HTTP = "HTTP"
    WEBSOCKETS = "WEBSOCKETS"
    COAP = "COAP"
    MQTT = "MQTT"


class InteractionVerbs(EnumListMixin):
    """Interactions have one or more defined interaction verbs for each
    interaction pattern.  Form Relations allow an interaction to have
    separate protocol mechanisms to support different interaction verbs."""

    READ_PROPERTY = "readproperty"
    WRITE_PROPERTY = "writeproperty"
    OBSERVE_PROPERTY = "observeproperty"
    INVOKE_ACTION = "invokeaction"
    SUBSCRIBE_EVENT = "subscribeevent"
    UNSUBSCRIBE_EVENT = "unsubscribeevent"



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import random
import uuid

import six
import tornado.concurrent
import tornado.gen
import tornado.ioloop
from faker import Faker
from rx.concurrency import IOLoopScheduler

from tests.utils import run_test_coroutine
from wotpy.wot.dictionaries.interaction import PropertyFragmentDict, EventFragmentDict, ActionFragmentDict
from wotpy.wot.td import ThingDescription


def client_test_on_property_change(servient, protocol_client_cls):
    """Helper function to test observation of Property updates on bindings clients."""

    exposed_thing = next(servient.exposed_things)

    prop_name = uuid.uuid4().hex

    exposed_thing.add_property(prop_name, PropertyFragmentDict({
        "type": "string",
        "observable": True
    }), value=Faker().sentence())

    servient.refresh_forms()

    td = ThingDescription.from_thing(exposed_thing.thing)

    @tornado.gen.coroutine
    def test_coroutine():
        protocol_client = protocol_client_cls()

        values = [Faker().sentence() for _ in range(10)]
        values_observed = {value: tornado.concurrent.Future() for value in values}

        @tornado.gen.coroutine
        def write_next():
            try:
                next_value = next(val for val, fut in six.iteritems(values_observed) if not fut.done())
                yield exposed_thing.properties[prop_name].write(next_value)
            except StopIteration:
                pass

        def on_next(ev):
            prop_value = ev.data.value
            if prop_value in values_observed and not values_observed[prop_value].done():
                values_observed[prop_value].set_result(True)

        observable = protocol_client.on_property_change(td, prop_name)

        subscription = observable.subscribe_on(IOLoopScheduler()).subscribe(on_next)

        periodic_emit = tornado.ioloop.PeriodicCallback(write_next, 10)
        periodic_emit.start()

        yield list(values_observed.values())

        periodic_emit.stop()
        subscription.dispose()

    run_test_coroutine(test_coroutine)


def client_test_on_event(servient, protocol_client_cls):
    """Helper function to test observation of Events on bindings clients."""

    exposed_thing = next(servient.exposed_things)

    event_name = uuid.uuid4().hex

    exposed_thing.add_event(event_name, EventFragmentDict({
        "type": "number"
    }))

    servient.refresh_forms()

    td = ThingDescription.from_thing(exposed_thing.thing)

    @tornado.gen.coroutine
    def test_coroutine():
        protocol_client = protocol_client_cls()

        payloads = [Faker().pyint() for _ in range(10)]
        future_payloads = {key: tornado.concurrent.Future() for key in payloads}

        @tornado.gen.coroutine
        def emit_next():
            next_value = next(val for val, fut in six.iteritems(future_payloads) if not fut.done())
            exposed_thing.events[event_name].emit(next_value)

        def on_next(ev):
            if ev.data in future_payloads and not future_payloads[ev.data].done():
                future_payloads[ev.data].set_result(True)

        observable = protocol_client.on_event(td, event_name)

        subscription = observable.subscribe_on(IOLoopScheduler()).subscribe(on_next)

        periodic_emit = tornado.ioloop.PeriodicCallback(emit_next, 10)
        periodic_emit.start()

        yield list(future_payloads.values())

        periodic_emit.stop()
        subscription.dispose()

    run_test_coroutine(test_coroutine)


def client_test_read_property(servient, protocol_client_cls, timeout=None):
    """Helper function to test Property reads on bindings clients."""

    exposed_thing = next(servient.exposed_things)

    prop_name = uuid.uuid4().hex

    exposed_thing.add_property(prop_name, PropertyFragmentDict({
        "type": "string",
        "observable": True
    }), value=Faker().sentence())

    servient.refresh_forms()

    td = ThingDescription.from_thing(exposed_thing.thing)

    @tornado.gen.coroutine
    def test_coroutine():
        protocol_client = protocol_client_cls()
        prop_value = Faker().sentence()

        curr_prop_value = yield protocol_client.read_property(td, prop_name, timeout=timeout)

        assert curr_prop_value != prop_value

        yield exposed_thing.properties[prop_name].write(prop_value)

        curr_prop_value = yield protocol_client.read_property(td, prop_name, timeout=timeout)

        assert curr_prop_value == prop_value

    run_test_coroutine(test_coroutine)


def client_test_write_property(servient, protocol_client_cls, timeout=None):
    """Helper function to test Property writes on bindings clients."""

    exposed_thing = next(servient.exposed_things)

    prop_name = uuid.uuid4().hex

    exposed_thing.add_property(prop_name, PropertyFragmentDict({
        "type": "string",
        "observable": True
    }), value=Faker().sentence())

    servient.refresh_forms()

    td = ThingDescription.from_thing(exposed_thing.thing)

    @tornado.gen.coroutine
    def test_coroutine():
        protocol_client = protocol_client_cls()
        prop_value = Faker().sentence()

        prev_value = yield exposed_thing.properties[prop_name].read()
        assert prev_value != prop_value

        yield protocol_client.write_property(td, prop_name, prop_value, timeout=timeout)

        curr_value = yield exposed_thing.properties[prop_name].read()
        assert curr_value == prop_value

    run_test_coroutine(test_coroutine)


def client_test_invoke_action(servient, protocol_client_cls, timeout=None):
    """Helper function to test Action invocations on bindings clients."""

    exposed_thing = next(servient.exposed_things)

    action_name = uuid.uuid4().hex

    @tornado.gen.coroutine
    def action_handler(parameters):
        input_value = parameters.get("input")
        yield tornado.gen.sleep(random.random() * 0.1)
        raise tornado.gen.Return("{:f}".format(input_value))

    exposed_thing.add_action(action_name, ActionFragmentDict({
        "input": {"type": "number"},
        "output": {"type": "string"}
    }), action_handler)

    servient.refresh_forms()

    td = ThingDescription.from_thing(exposed_thing.thing)

    @tornado.gen.coroutine
    def test_coroutine():
        protocol_client = protocol_client_cls()

        input_value = Faker().pyint()

        result = yield protocol_client.invoke_action(td, action_name, input_value, timeout=timeout)
        result_expected = yield action_handler({"input": input_value})

        assert result == result_expected

    run_test_coroutine(test_coroutine)


def client_test_invoke_action_error(servient, protocol_client_cls):
    """Helper function to test Action invocations that raise errors on bindings clients."""

    exposed_thing = next(servient.exposed_things)

    action_name = uuid.uuid4().hex

    err_message = Faker().sentence()

    # noinspection PyUnusedLocal
    def action_handler(parameters):
        raise ValueError(err_message)

    exposed_thing.add_action(action_name, ActionFragmentDict({
        "input": {"type": "number"},
        "output": {"type": "string"}
    }), action_handler)

    servient.refresh_forms()

    td = ThingDescription.from_thing(exposed_thing.thing)

    @tornado.gen.coroutine
    def test_coroutine():
        protocol_client = protocol_client_cls()

        try:
            yield protocol_client.invoke_action(td, action_name, Faker().pyint())
            raise AssertionError("Did not raise Exception")
        except Exception as ex:
            assert err_message in str(ex)

    run_test_coroutine(test_coroutine)


def client_test_on_property_change_error(servient, protocol_client_cls):
    """Helper function to test propagation of errors raised
    during observation of Property updates on bindings clients."""

    exposed_thing = next(servient.exposed_things)

    prop_name = uuid.uuid4().hex

    exposed_thing.add_property(prop_name, PropertyFragmentDict({
        "type": "string",
        "observable": True
    }), value=Faker().sentence())

    servient.refresh_forms()

    td = ThingDescription.from_thing(exposed_thing.thing)

    @tornado.gen.coroutine
    def test_coroutine():
        protocol_client = protocol_client_cls()

        yield servient.shutdown()

        future_err = tornado.concurrent.Future()

        # noinspection PyUnusedLocal
        def on_next(item):
            future_err.set_exception(Exception("Should not have emitted any items"))

        def on_error(err):
            future_err.set_result(err)

        observable = protocol_client.on_property_change(td, prop_name)

        subscribe_kwargs = {
            "on_next": on_next,
            "on_error": on_error
        }

        subscription = observable.subscribe_on(IOLoopScheduler()).subscribe(**subscribe_kwargs)

        observe_err = yield future_err

        assert isinstance(observe_err, Exception)

        subscription.dispose()

    run_test_coroutine(test_coroutine)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
WebSockets Protocol Binding implementation.

.. autosummary::
    :toctree: _ws

    wotpy.protocols.ws.client
    wotpy.protocols.ws.enums
    wotpy.protocols.ws.handler
    wotpy.protocols.ws.messages
    wotpy.protocols.ws.schemas
    wotpy.protocols.ws.server
"""



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
HTTP Protocol Binding implementation.

.. autosummary::
    :toctree: _http

    wotpy.protocols.http.handlers
    wotpy.protocols.http.client
    wotpy.protocols.http.enums
    wotpy.protocols.http.server
"""



#!/usr/bin/env python
# -*- coding: utf-8 -*-


def assert_exposed_thing_equal(exp_thing, td_doc):
    """Asserts that the given ExposedThing is equivalent to the Thing Description dict."""

    assert exp_thing.thing.id == td_doc.get("id")
    assert exp_thing.thing.name == td_doc.get("name", None)
    assert exp_thing.thing.description == td_doc.get("description", None)
    assert len(exp_thing.thing.properties) == len(td_doc.get("properties", []))
    assert len(exp_thing.thing.actions) == len(td_doc.get("actions", []))
    assert len(exp_thing.thing.events) == len(td_doc.get("events", []))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
MQTT handler for PING requests published on the MQTT broker.
"""

import tornado.gen
from hbmqtt.mqtt.constants import QOS_1

from wotpy.protocols.mqtt.handlers.base import BaseMQTTHandler


class PingMQTTHandler(BaseMQTTHandler):
    """MQTT handler for PING requests published on the MQTT broker."""

    def __init__(self, mqtt_server, qos=QOS_1):
        super(PingMQTTHandler, self).__init__(mqtt_server)
        self._qos = qos

    @property
    def topic_ping(self):
        """Ping topic."""

        return "{}/ping".format(self.servient_id)

    @property
    def topic_pong(self):
        """Pong topic."""

        return "{}/pong".format(self.servient_id)

    @property
    def topics(self):
        """List of topics that this MQTT handler wants to subscribe to."""

        return [(self.topic_ping, self._qos)]

    @tornado.gen.coroutine
    def handle_message(self, msg):
        """Publishes a message in the PONG topic with the
        same payload as the one received in the PING topic."""

        yield self.queue.put({
            "topic": self.topic_pong,
            "data": msg.data,
            "qos": self._qos
        })



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
MQTT handler for Event subscriptions.
"""

import json
import time

import tornado.gen
import tornado.ioloop
from hbmqtt.mqtt.constants import QOS_0
from tornado.queues import QueueFull

from wotpy.protocols.mqtt.handlers.base import BaseMQTTHandler
from wotpy.protocols.mqtt.handlers.subs import InteractionsSubscriber
from wotpy.utils.utils import to_json_obj
from wotpy.wot.enums import InteractionTypes


class EventMQTTHandler(BaseMQTTHandler):
    """MQTT handler for Event subscriptions."""

    DEFAULT_CALLBACK_MS = 2000
    DEFAULT_JITTER = 0.2

    def __init__(self, mqtt_server, qos=QOS_0, callback_ms=None):
        super(EventMQTTHandler, self).__init__(mqtt_server)

        callback_ms = self.DEFAULT_CALLBACK_MS if callback_ms is None else callback_ms

        self._qos = qos
        self._callback_ms = callback_ms
        self._subs = {}

        self._interaction_subscriber = InteractionsSubscriber(
            interaction_type=InteractionTypes.EVENT,
            server=self.mqtt_server,
            on_next_builder=self._build_on_next)

        @tornado.gen.coroutine
        def refresh_subs():
            self._interaction_subscriber.refresh()

        self._periodic_refresh_subs = tornado.ioloop.PeriodicCallback(
            refresh_subs, self._callback_ms, jitter=self.DEFAULT_JITTER)

    def build_event_topic(self, thing, event):
        """Returns the MQTT topic for Event emissions."""

        return "{}/event/{}/{}".format(
            self.servient_id,
            thing.url_name,
            event.url_name)

    @tornado.gen.coroutine
    def init(self):
        """Initializes the MQTT handler.
        Called when the MQTT runner starts."""

        self._interaction_subscriber.refresh()
        self._periodic_refresh_subs.start()

        yield None

    @tornado.gen.coroutine
    def teardown(self):
        """Destroys the MQTT handler.
        Called when the MQTT runner stops."""

        self._periodic_refresh_subs.stop()
        self._interaction_subscriber.dispose()

        yield None

    def _build_on_next(self, exp_thing, event):
        """Builds the on_next function to use when subscribing to the given Event."""

        topic = self.build_event_topic(exp_thing, event)

        def on_next(item):
            try:
                data = {
                    "name": item.name,
                    "data": to_json_obj(item.data),
                    "timestamp": int(time.time() * 1000)
                }

                self.queue.put_nowait({
                    "topic": topic,
                    "data": json.dumps(data).encode(),
                    "qos": self._qos
                })
            except QueueFull:
                pass

        return on_next



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Schemas following the JSON Schema specification used to validate the shape of WebSockets messages.
"""

from wotpy.protocols.ws.enums import WebsocketMethods

JSON_RPC_VERSION = "2.0"

# Schema for message IDs

SCHEMA_ID = {
    "oneOf": [
        {"type": "string"},
        {"type": "integer"},
        {"type": "null"}
    ]
}

SCHEMA_REQUEST = {
    "$schema": "http://json-schema.org/schema#",
    "id": "http://fundacionctic.org/schemas/wotpy-ws-request.json",
    "type": "object",
    "properties": {
        "jsonrpc": {
            "type": "string",
            "enum": [JSON_RPC_VERSION]
        },
        "method": {
            "type": "string",
            "enum": WebsocketMethods.list()
        },
        "params": {
            "oneOf": [
                {"type": "object"},
                {"type": "array"}
            ]
        },
        "id": SCHEMA_ID
    },
    "required": [
        "jsonrpc",
        "method"
    ]
}

SCHEMA_RESPONSE = {
    "$schema": "http://json-schema.org/schema#",
    "id": "http://fundacionctic.org/schemas/wotpy-ws-response.json",
    "type": "object",
    "properties": {
        "jsonrpc": {
            "type": "string",
            "enum": ["2.0"]
        },
        "result": {},
        "id": SCHEMA_ID
    },
    "required": [
        "jsonrpc",
        "result",
        "id"
    ]
}

SCHEMA_ERROR = {
    "$schema": "http://json-schema.org/schema#",
    "id": "http://fundacionctic.org/schemas/wotpy-ws-error.json",
    "type": "object",
    "properties": {
        "jsonrpc": {
            "type": "string",
            "enum": ["2.0"]
        },
        "error": {
            "type": "object",
            "properties": {
                "code": {"type": "integer"},
                "message": {"type": "string"},
                "data": {}
            },
            "required": [
                "code",
                "message"
            ]
        },
        "id": SCHEMA_ID
    },
    "required": [
        "jsonrpc",
        "error",
        "id"
    ]
}

SCHEMA_EMITTED_ITEM = {
    "$schema": "http://json-schema.org/schema#",
    "id": "http://fundacionctic.org/schemas/wotpy-ws-emitted-item.json",
    "type": "object",
    "properties": {
        "subscription": {"type": "string"},
        "name": {"type": "string"},
        "data": {}
    },
    "required": [
        "subscription",
        "name",
        "data"
    ]
}

SCHEMA_PARAMS_READ_PROPERTY = {
    "$schema": "http://json-schema.org/schema#",
    "id": "http://fundacionctic.org/schemas/wotpy-ws-params-read-property.json",
    "type": "object",
    "properties": {
        "name": {"type": "string"}
    },
    "required": [
        "name"
    ]
}

SCHEMA_PARAMS_WRITE_PROPERTY = {
    "$schema": "http://json-schema.org/schema#",
    "id": "http://fundacionctic.org/schemas/wotpy-ws-params-write-property.json",
    "type": "object",
    "properties": {
        "name": {"type": "string"},
        "value": {}
    },
    "required": [
        "name",
        "value"
    ]
}

SCHEMA_PARAMS_INVOKE_ACTION = {
    "$schema": "http://json-schema.org/schema#",
    "id": "http://fundacionctic.org/schemas/wotpy-ws-params-invoke-action.json",
    "type": "object",
    "properties": {
        "name": {"type": "string"},
        "parameters": {}
    },
    "required": [
        "name"
    ]
}

SCHEMA_PARAMS_ON_PROPERTY_CHANGE = {
    "$schema": "http://json-schema.org/schema#",
    "id": "http://fundacionctic.org/schemas/wotpy-ws-params-on-property-change.json",
    "type": "object",
    "properties": {
        "name": {"type": "string"}
    },
    "required": [
        "name"
    ]
}

SCHEMA_PARAMS_ON_TD_CHANGE = {
    "$schema": "http://json-schema.org/schema#",
    "id": "http://fundacionctic.org/schemas/wotpy-ws-params-on-td-change.json",
    "type": "object"
}

SCHEMA_PARAMS_ON_EVENT = {
    "$schema": "http://json-schema.org/schema#",
    "id": "http://fundacionctic.org/schemas/wotpy-ws-params-on-event.json",
    "type": "object",
    "properties": {
        "name": {"type": "string"}
    },
    "required": [
        "name"
    ]
}

SCHEMA_PARAMS_DISPOSE = {
    "$schema": "http://json-schema.org/schema#",
    "id": "http://fundacionctic.org/schemas/wotpy-ws-params-dispose.json",
    "type": "object",
    "properties": {
        "subscription": {"type": "string"}
    },
    "required": [
        "subscription"
    ]
}



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import json
import time
import uuid
from asyncio import TimeoutError

import pytest
import six
import tornado.concurrent
import tornado.gen
import tornado.ioloop
from faker import Faker
from hbmqtt.client import MQTTClient
from hbmqtt.mqtt.constants import QOS_2, QOS_0

from tests.protocols.mqtt.broker import is_test_broker_online, BROKER_SKIP_REASON, get_test_broker_url
from tests.utils import run_test_coroutine
from wotpy.protocols.enums import InteractionVerbs
from wotpy.protocols.mqtt.handlers.action import ActionMQTTHandler
from wotpy.protocols.mqtt.server import MQTTServer
from wotpy.wot.dictionaries.interaction import PropertyFragmentDict, ActionFragmentDict

pytestmark = pytest.mark.skipif(is_test_broker_online() is False, reason=BROKER_SKIP_REASON)


def build_topic(server, interaction, interaction_verb):
    """Returns the topic for the given interaction and verb."""

    forms = server.build_forms(None, interaction)
    form = next(item for item in forms if interaction_verb == item.op)
    return "/".join(form.href.split("/")[3:])


@tornado.gen.coroutine
def connect_broker(topics):
    """Connects to the test MQTT broker and subscribes to the topics."""

    topics = [(topics, QOS_0)] if isinstance(topics, six.string_types) else topics

    hbmqtt_client = MQTTClient()
    yield hbmqtt_client.connect(get_test_broker_url())
    yield hbmqtt_client.subscribe(topics)

    raise tornado.gen.Return(hbmqtt_client)


@tornado.gen.coroutine
def _ping(mqtt_server, timeout=None):
    """Returns True if the given MQTT server has answered to a PING request."""

    broker_url = get_test_broker_url()

    topic_ping = "{}/ping".format(mqtt_server.servient_id)
    topic_pong = "{}/pong".format(mqtt_server.servient_id)

    try:
        hbmqtt_client = MQTTClient()
        yield hbmqtt_client.connect(broker_url)
        yield hbmqtt_client.subscribe([(topic_pong, QOS_2)])
        bytes_payload = bytes(uuid.uuid4().hex, "utf8")
        yield hbmqtt_client.publish(topic_ping, bytes_payload, qos=QOS_2)
        message = yield hbmqtt_client.deliver_message(timeout=timeout)
        assert message.data == bytes_payload
        yield hbmqtt_client.disconnect()
    except TimeoutError:
        raise tornado.gen.Return(False)

    raise tornado.gen.Return(True)


DEFAULT_PING_TIMEOUT = 1.0


def test_start_stop():
    """The MQTT server may be started and stopped."""

    mqtt_server = MQTTServer(broker_url=get_test_broker_url())

    @tornado.gen.coroutine
    def test_coroutine():
        assert not (yield _ping(mqtt_server, timeout=DEFAULT_PING_TIMEOUT))

        yield mqtt_server.start()

        assert (yield _ping(mqtt_server))
        assert (yield _ping(mqtt_server))

        yield mqtt_server.stop()
        yield mqtt_server.start()
        yield mqtt_server.stop()

        assert not (yield _ping(mqtt_server, timeout=DEFAULT_PING_TIMEOUT))

        yield mqtt_server.stop()
        yield mqtt_server.start()
        yield mqtt_server.start()

        assert (yield _ping(mqtt_server))

    run_test_coroutine(test_coroutine)


def test_servient_id():
    """An MQTT server may be identified by a unique Servient ID to avoid topic collisions."""

    broker_url = get_test_broker_url()

    mqtt_srv_01 = MQTTServer(broker_url=broker_url)
    mqtt_srv_02 = MQTTServer(broker_url=broker_url)
    mqtt_srv_03 = MQTTServer(broker_url=broker_url, servient_id=Faker().pystr())

    assert mqtt_srv_01.servient_id and mqtt_srv_02.servient_id and mqtt_srv_03.servient_id
    assert mqtt_srv_01.servient_id == mqtt_srv_02.servient_id
    assert mqtt_srv_01.servient_id != mqtt_srv_03.servient_id

    @tornado.gen.coroutine
    def assert_ping_loop(srv, num_iters=10):
        for _ in range(num_iters):
            assert (yield _ping(srv, timeout=DEFAULT_PING_TIMEOUT))

    @tornado.gen.coroutine
    def test_coroutine():
        yield [
            mqtt_srv_01.start(),
            mqtt_srv_03.start()
        ]

        yield [
            assert_ping_loop(mqtt_srv_01),
            assert_ping_loop(mqtt_srv_03)
        ]

    run_test_coroutine(test_coroutine)


def test_property_read(mqtt_server):
    """Current Property values may be requested using the MQTT binding."""

    exposed_thing = next(mqtt_server.exposed_things)
    prop_name = next(six.iterkeys(exposed_thing.thing.properties))
    prop = exposed_thing.thing.properties[prop_name]
    topic_read = build_topic(mqtt_server, prop, InteractionVerbs.READ_PROPERTY)
    topic_observe = build_topic(mqtt_server, prop, InteractionVerbs.OBSERVE_PROPERTY)

    observe_timeout_secs = 1.0

    @tornado.gen.coroutine
    def test_coroutine():
        prop_value = yield exposed_thing.properties[prop_name].read()

        client_read = yield connect_broker(topic_read)
        client_observe = yield connect_broker(topic_observe)

        try:
            yield client_observe.deliver_message(timeout=observe_timeout_secs)
            raise AssertionError('Unexpected message on topic {}'.format(topic_observe))
        except TimeoutError:
            pass

        @tornado.gen.coroutine
        def read_value():
            payload = json.dumps({"action": "read"}).encode()
            yield client_read.publish(topic_read, payload, qos=QOS_2)

        periodic_read = tornado.ioloop.PeriodicCallback(read_value, 50)
        periodic_read.start()

        msg = yield client_observe.deliver_message()

        periodic_read.stop()

        assert json.loads(msg.data.decode()).get("value") == prop_value

    run_test_coroutine(test_coroutine)


def test_property_write(mqtt_server):
    """Property values may be updated using the MQTT binding."""

    exposed_thing = next(mqtt_server.exposed_things)
    prop_name = next(six.iterkeys(exposed_thing.thing.properties))
    prop = exposed_thing.thing.properties[prop_name]
    topic_write = build_topic(mqtt_server, prop, InteractionVerbs.WRITE_PROPERTY)
    topic_observe = build_topic(mqtt_server, prop, InteractionVerbs.OBSERVE_PROPERTY)

    @tornado.gen.coroutine
    def test_coroutine():
        updated_value = Faker().sentence()

        client_write = yield connect_broker(topic_write)
        client_observe = yield connect_broker(topic_observe)

        future_observe = tornado.concurrent.Future()

        @tornado.gen.coroutine
        def resolve_future_on_update():
            msg_observe = yield client_observe.deliver_message()
            assert json.loads(msg_observe.data.decode()).get("value") == updated_value
            future_observe.set_result(True)

        tornado.ioloop.IOLoop.current().spawn_callback(resolve_future_on_update)

        @tornado.gen.coroutine
        def publish_write():
            payload = json.dumps({"action": "write", "value": updated_value}).encode()
            yield client_write.publish(topic_write, payload, qos=QOS_2)

        periodic_write = tornado.ioloop.PeriodicCallback(publish_write, 50)
        periodic_write.start()

        yield future_observe

        assert future_observe.result() is True

        periodic_write.stop()

    run_test_coroutine(test_coroutine)


CALLBACK_MS = 50


@pytest.mark.parametrize("mqtt_server", [{"property_callback_ms": CALLBACK_MS}], indirect=True)
def test_property_add_remove(mqtt_server):
    """The MQTT binding reacts appropriately to Properties
    being added and removed from ExposedThings."""

    exposed_thing = next(mqtt_server.exposed_things)
    prop_names = list(six.iterkeys(exposed_thing.thing.properties))

    for name in prop_names:
        exposed_thing.remove_property(name)

    def add_prop(pname):
        exposed_thing.add_property(pname, PropertyFragmentDict({
            "type": "number",
            "observable": True
        }), value=Faker().pyint())

    def del_prop(pname):
        exposed_thing.remove_property(pname)

    @tornado.gen.coroutine
    def is_prop_active(prop, timeout_secs=1.0):
        topic_observe = build_topic(mqtt_server, prop, InteractionVerbs.OBSERVE_PROPERTY)
        topic_write = build_topic(mqtt_server, prop, InteractionVerbs.WRITE_PROPERTY)

        client_observe = yield connect_broker(topic_observe)
        client_write = yield connect_broker(topic_write)

        value = Faker().pyint()

        @tornado.gen.coroutine
        def publish_write():
            payload = json.dumps({"action": "write", "value": value}).encode()
            yield client_write.publish(topic_write, payload, qos=QOS_0)

        write_interval = (timeout_secs / 4.0) * 1000.0
        periodic_write = tornado.ioloop.PeriodicCallback(publish_write, write_interval)
        periodic_write.start()

        try:
            msg = yield client_observe.deliver_message(timeout=timeout_secs)
            assert json.loads(msg.data.decode()).get("value") == value
            raise tornado.gen.Return(True)
        except TimeoutError:
            raise tornado.gen.Return(False)
        finally:
            periodic_write.stop()

    @tornado.gen.coroutine
    def test_coroutine():
        sleep_secs = (CALLBACK_MS / 1000.0) * 4

        prop_01_name = uuid.uuid4().hex
        prop_02_name = uuid.uuid4().hex
        prop_03_name = uuid.uuid4().hex

        add_prop(prop_01_name)
        add_prop(prop_02_name)
        add_prop(prop_03_name)

        prop_01 = exposed_thing.thing.properties[prop_01_name]
        prop_02 = exposed_thing.thing.properties[prop_02_name]
        prop_03 = exposed_thing.thing.properties[prop_03_name]

        yield tornado.gen.sleep(sleep_secs)

        assert (yield is_prop_active(prop_01))
        assert (yield is_prop_active(prop_02))
        assert (yield is_prop_active(prop_03))

        del_prop(prop_01_name)

        assert not (yield is_prop_active(prop_01))
        assert (yield is_prop_active(prop_02))
        assert (yield is_prop_active(prop_03))

        del_prop(prop_03_name)

        assert not (yield is_prop_active(prop_01))
        assert (yield is_prop_active(prop_02))
        assert not (yield is_prop_active(prop_03))

        add_prop(prop_01_name)

        prop_01 = exposed_thing.thing.properties[prop_01_name]

        assert (yield is_prop_active(prop_01))
        assert (yield is_prop_active(prop_02))
        assert not (yield is_prop_active(prop_03))

    run_test_coroutine(test_coroutine)


def test_observe_property_changes(mqtt_server):
    """Property updates may be observed using the MQTT binding."""

    exposed_thing = next(mqtt_server.exposed_things)
    prop_name = next(six.iterkeys(exposed_thing.thing.properties))
    prop = exposed_thing.thing.properties[prop_name]
    topic_observe = build_topic(mqtt_server, prop, InteractionVerbs.OBSERVE_PROPERTY)

    @tornado.gen.coroutine
    def test_coroutine():
        client_observe = yield connect_broker(topic_observe)

        updated_value = Faker().sentence()

        @tornado.gen.coroutine
        def write_value():
            yield exposed_thing.properties[prop_name].write(updated_value)

        periodic_write = tornado.ioloop.PeriodicCallback(write_value, 50)
        periodic_write.start()

        msg = yield client_observe.deliver_message()

        assert json.loads(msg.data.decode()).get("value") == updated_value

        periodic_write.stop()

    run_test_coroutine(test_coroutine)


def test_observe_event(mqtt_server):
    """Events may be observed using the MQTT binding."""

    now_ms = int(time.time() * 1000)

    exposed_thing = next(mqtt_server.exposed_things)
    event_name = next(six.iterkeys(exposed_thing.thing.events))
    event = exposed_thing.thing.events[event_name]
    topic = build_topic(mqtt_server, event, InteractionVerbs.SUBSCRIBE_EVENT)

    @tornado.gen.coroutine
    def test_coroutine():
        client = yield connect_broker(topic)

        emitted_value = Faker().pyint()

        @tornado.gen.coroutine
        def emit_value():
            yield exposed_thing.events[event_name].emit(emitted_value)

        periodic_emit = tornado.ioloop.PeriodicCallback(emit_value, 50)
        periodic_emit.start()

        msg = yield client.deliver_message()

        event_data = json.loads(msg.data.decode())

        assert event_data.get("name") == event_name
        assert event_data.get("data") == emitted_value
        assert event_data.get("timestamp") >= now_ms

        periodic_emit.stop()

    run_test_coroutine(test_coroutine)


def test_action_invoke(mqtt_server):
    """Actions can be invoked using the MQTT binding."""

    exposed_thing = next(mqtt_server.exposed_things)
    action_name = next(six.iterkeys(exposed_thing.thing.actions))
    action = exposed_thing.thing.actions[action_name]

    topic_invoke = build_topic(mqtt_server, action, InteractionVerbs.INVOKE_ACTION)
    topic_result = ActionMQTTHandler.to_result_topic(topic_invoke)

    @tornado.gen.coroutine
    def test_coroutine():
        client_invoke = yield connect_broker(topic_invoke)
        client_result = yield connect_broker(topic_result)

        data = {
            "id": uuid.uuid4().hex,
            "input": Faker().pyint()
        }

        now_ms = int(time.time() * 1000)

        yield client_invoke.publish(topic_invoke, json.dumps(data).encode(), qos=QOS_2)

        msg = yield client_result.deliver_message()
        msg_data = json.loads(msg.data.decode())

        assert msg_data.get("id") == data.get("id")
        assert msg_data.get("result") == "{:f}".format(data.get("input"))
        assert msg_data.get("timestamp") >= now_ms

    run_test_coroutine(test_coroutine)


def test_action_invoke_error(mqtt_server):
    """Action errors are handled appropriately by the MQTT binding."""

    exposed_thing = next(mqtt_server.exposed_things)

    action_name = uuid.uuid4().hex
    err_message = Faker().sentence()

    # noinspection PyUnusedLocal
    def handler(parameters):
        raise TypeError(err_message)

    exposed_thing.add_action(action_name, ActionFragmentDict({
        "input": {"type": "string"},
        "output": {"type": "string"}
    }), handler)

    action = exposed_thing.thing.actions[action_name]

    topic_invoke = build_topic(mqtt_server, action, InteractionVerbs.INVOKE_ACTION)
    topic_result = ActionMQTTHandler.to_result_topic(topic_invoke)

    @tornado.gen.coroutine
    def test_coroutine():
        client_invoke = yield connect_broker(topic_invoke)
        client_result = yield connect_broker(topic_result)

        data = {
            "id": uuid.uuid4().hex,
            "input": Faker().pyint()
        }

        yield client_invoke.publish(topic_invoke, json.dumps(data).encode(), qos=QOS_2)

        msg = yield client_result.deliver_message()
        msg_data = json.loads(msg.data.decode())

        assert msg_data.get("id") == data.get("id")
        assert msg_data.get("error") == err_message
        assert msg_data.get("result", None) is None

    run_test_coroutine(test_coroutine)


def test_action_invoke_parallel(mqtt_server):
    """Multiple Actions can be invoked in parallel using the MQTT binding."""

    exposed_thing = next(mqtt_server.exposed_things)
    action_name = next(six.iterkeys(exposed_thing.thing.actions))
    action = exposed_thing.thing.actions[action_name]

    topic_invoke = build_topic(mqtt_server, action, InteractionVerbs.INVOKE_ACTION)
    topic_result = ActionMQTTHandler.to_result_topic(topic_invoke)

    num_requests = 10

    @tornado.gen.coroutine
    def test_coroutine():
        client_invoke = yield connect_broker(topic_invoke)
        client_result = yield connect_broker(topic_result)

        requests = []

        for idx in range(num_requests):
            requests.append({
                "id": uuid.uuid4().hex,
                "input": Faker().pyint()
            })

        now_ms = int(time.time() * 1000)

        yield [
            client_invoke.publish(topic_invoke, json.dumps(requests[idx]).encode(), qos=QOS_2)
            for idx in range(num_requests)
        ]

        for idx in range(num_requests):
            msg = yield client_result.deliver_message()
            msg_data = json.loads(msg.data.decode())
            expected = next(item for item in requests if item.get("id") == msg_data.get("id"))

            assert msg_data.get("id") == expected.get("id")
            assert msg_data.get("result") == "{:f}".format(expected.get("input"))
            assert msg_data.get("timestamp") >= now_ms

    run_test_coroutine(test_coroutine)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
MQTT handler for Action invocations.
"""

import json
import time
from json import JSONDecodeError

import tornado.gen
import tornado.ioloop
from hbmqtt.mqtt.constants import QOS_2

from wotpy.protocols.mqtt.handlers.base import BaseMQTTHandler
from wotpy.utils.utils import to_json_obj


class ActionMQTTHandler(BaseMQTTHandler):
    """MQTT handler for Action invocations."""

    KEY_INPUT = "input"
    KEY_INVOCATION_ID = "id"

    def __init__(self, mqtt_server, qos=QOS_2):
        super(ActionMQTTHandler, self).__init__(mqtt_server)

        self._qos = qos

    @property
    def topic_wildcard_invocation(self):
        """Wildcard topic to subscribe to all Action invocations."""

        return "{}/action/invocation/#".format(self.servient_id)

    @classmethod
    def to_result_topic(cls, invocation_topic):
        """Takes an Action invocation MQTT topic and returns the related result topic."""

        topic_split = invocation_topic.split("/")
        servient_id, thing_name, action_name = topic_split[-5], topic_split[-2], topic_split[-1]

        return "{}/action/result/{}/{}".format(
            servient_id,
            thing_name,
            action_name)

    def build_action_result_topic(self, thing, action):
        """Returns the MQTT topic for Action invocation results."""

        return "{}/action/result/{}/{}".format(
            self.servient_id,
            thing.url_name,
            action.url_name)

    @property
    def topics(self):
        """List of topics that this MQTT handler wants to subscribe to."""

        return [(self.topic_wildcard_invocation, self._qos)]

    @tornado.gen.coroutine
    def handle_message(self, msg):
        """Listens to all Property request topics and responds to read and write requests."""

        now_ms = int(time.time() * 1000)

        try:
            parsed_msg = json.loads(msg.data.decode())
        except (JSONDecodeError, TypeError):
            return

        topic_split = msg.topic.split("/")

        splits_expected_len = len(self.topic_wildcard_invocation.split("/")) + 1

        if len(topic_split) != splits_expected_len:
            return

        thing_url_name, action_url_name = topic_split[-2], topic_split[-1]

        try:
            exp_thing = next(
                item for item in self.mqtt_server.exposed_things
                if item.url_name == thing_url_name)

            action = next(
                exp_thing.thing.actions[key] for key in exp_thing.thing.actions
                if exp_thing.thing.actions[key].url_name == action_url_name)
        except StopIteration:
            return

        input_value = parsed_msg.get(self.KEY_INPUT, None)

        data = {
            "id": parsed_msg.get(self.KEY_INVOCATION_ID, None),
            "timestamp": now_ms
        }

        try:
            result = yield exp_thing.actions[action.name].invoke(input_value)
            data.update({"result": to_json_obj(result)})
        except Exception as ex:
            data.update({"error": str(ex)})

        topic = self.build_action_result_topic(exp_thing.thing, action)

        yield self.queue.put({
            "topic": topic,
            "data": json.dumps(data).encode(),
            "qos": self._qos
        })



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from tests.protocols.helpers import \
    client_test_on_property_change, \
    client_test_on_event, \
    client_test_read_property, \
    client_test_write_property, \
    client_test_invoke_action, \
    client_test_invoke_action_error, \
    client_test_on_property_change_error
from wotpy.protocols.coap.client import CoAPClient


def test_read_property(coap_servient):
    """The CoAP client can read properties."""

    client_test_read_property(coap_servient, CoAPClient)


def test_write_property(coap_servient):
    """The CoAP client can write properties."""

    client_test_write_property(coap_servient, CoAPClient)


def test_on_property_change(coap_servient):
    """The CoAP client can subscribe to property updates."""

    client_test_on_property_change(coap_servient, CoAPClient)


def test_on_property_change_error(coap_servient):
    """Errors that arise in the middle of an ongoing Property
    observation are propagated to the subscription as expected."""

    client_test_on_property_change_error(coap_servient, CoAPClient)


def test_invoke_action(coap_servient):
    """The CoAP client can invoke actions."""

    client_test_invoke_action(coap_servient, CoAPClient)


def test_invoke_action_error(coap_servient):
    """Errors raised by Actions are propagated propertly by the CoAP binding client."""

    client_test_invoke_action_error(coap_servient, CoAPClient)


def test_on_event(coap_servient):
    """The CoAP client can subscribe to event emissions."""

    client_test_on_event(coap_servient, CoAPClient)



#!/usr/bin/env python
# -*- coding: utf-8 -*-


import string
import uuid

# noinspection PyPackageRequirements
import pytest
# noinspection PyPackageRequirements
# noinspection PyPackageRequirements
from slugify import slugify

from wotpy.protocols.enums import Protocols
from wotpy.wot.td import ThingDescription
from wotpy.wot.form import Form
from wotpy.wot.interaction import Action
from wotpy.wot.thing import Thing


def test_unique_url_name():
    """URL names are always unique as long as the IDs are."""

    thing_id_base = uuid.uuid4().urn
    thing_ids = [thing_id_base]

    for ch in string.punctuation.replace("-", ""):
        thing_ids.append(thing_id_base.replace("-", ch))

    things = [Thing(id=item) for item in thing_ids]
    url_names = [item.url_name for item in things]

    assert len(url_names) == len(set(url_names))


def test_empty_thing_valid():
    """An empty Thing initialized by default has a valid JSON-LD serialization."""

    thing = Thing(id=uuid.uuid4().urn)
    json_td = ThingDescription.from_thing(thing)
    ThingDescription.validate(json_td.to_dict())


def test_interaction_invalid_name():
    """Invalid names for Interaction objects are rejected."""

    names_valid = [
        "safename",
        "safename02",
        "SafeName_03",
        "Safe_Name-04"
    ]

    names_invalid = [
        "!unsafename",
        "unsafe_name_ñ",
        "unsafe name",
        "?"
    ]

    thing = Thing(id=uuid.uuid4().urn)

    for name in names_valid:
        Action(thing=thing, name=name)

    for name in names_invalid:
        with pytest.raises(ValueError):
            Action(thing=thing, name=name)


def test_find_interaction():
    """Interactions may be retrieved by name on a Thing."""

    thing = Thing(id=uuid.uuid4().urn)

    interaction_01 = Action(thing=thing, name="my_interaction")
    interaction_02 = Action(thing=thing, name="AnotherInteraction")

    thing.add_interaction(interaction_01)
    thing.add_interaction(interaction_02)

    assert thing.find_interaction(interaction_01.name) is interaction_01
    assert thing.find_interaction(interaction_02.name) is interaction_02
    assert thing.find_interaction(slugify(interaction_01.name)) is interaction_01
    assert thing.find_interaction(slugify(interaction_02.name)) is interaction_02


def test_remove_interaction():
    """Interactions may be removed from a Thing by name."""

    thing = Thing(id=uuid.uuid4().urn)

    interaction_01 = Action(thing=thing, name="my_interaction")
    interaction_02 = Action(thing=thing, name="AnotherInteraction")
    interaction_03 = Action(thing=thing, name="YetAnother_interaction")

    thing.add_interaction(interaction_01)
    thing.add_interaction(interaction_02)
    thing.add_interaction(interaction_03)

    assert thing.find_interaction(interaction_01.name) is not None
    assert thing.find_interaction(interaction_02.name) is not None
    assert thing.find_interaction(interaction_03.name) is not None

    thing.remove_interaction(interaction_01.name)
    thing.remove_interaction(slugify(interaction_03.name))

    assert thing.find_interaction(interaction_01.name) is None
    assert thing.find_interaction(interaction_02.name) is not None
    assert thing.find_interaction(interaction_03.name) is None


def test_duplicated_interactions():
    """Duplicated Interactions are rejected on a Thing."""

    thing = Thing(id=uuid.uuid4().urn)

    interaction_01 = Action(thing=thing, name="my_interaction")
    interaction_02 = Action(thing=thing, name="AnotherInteraction")
    interaction_03 = Action(thing=thing, name="my_interaction")

    thing.add_interaction(interaction_01)
    thing.add_interaction(interaction_02)

    with pytest.raises(ValueError):
        thing.add_interaction(interaction_03)


def test_duplicated_forms():
    """Duplicated Forms are rejected on an Interaction."""

    thing = Thing(id=uuid.uuid4().urn)
    interaction = Action(thing=thing, name="my_interaction")
    thing.add_interaction(interaction)

    href_01 = "/href-01"
    href_02 = "/href-02"

    mtype_01 = "application/json"
    mtype_02 = "text/html"

    form_01 = Form(interaction=interaction, protocol=Protocols.HTTP, href=href_01, content_type=mtype_01)
    form_02 = Form(interaction=interaction, protocol=Protocols.HTTP, href=href_01, content_type=mtype_01)
    form_03 = Form(interaction=interaction, protocol=Protocols.HTTP, href=href_01, content_type=mtype_02)
    form_04 = Form(interaction=interaction, protocol=Protocols.HTTP, href=href_02, content_type=mtype_01)
    form_05 = Form(interaction=interaction, protocol=Protocols.HTTP, href=href_02, content_type=mtype_02)
    form_06 = Form(interaction=interaction, protocol=Protocols.HTTP, href=href_02, content_type=mtype_02)

    interaction.add_form(form_01)

    with pytest.raises(ValueError):
        interaction.add_form(form_02)

    interaction.add_form(form_03)
    interaction.add_form(form_04)
    interaction.add_form(form_05)

    with pytest.raises(ValueError):
        interaction.add_form(form_06)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Class that represents the form entities exposed by interactions.
"""

from wotpy.wot.dictionaries.link import FormDict


class Form(object):
    """Communication metadata where a service can be accessed by a client application."""

    def __init__(self, interaction, protocol, form_dict=None, **kwargs):
        self._interaction = interaction
        self._protocol = protocol
        self._form_dict = form_dict if form_dict else FormDict(**kwargs)

    def __getattr__(self, name):
        """Search for members that raised an AttributeError in
        the internal Form init dict before propagating the exception."""

        return getattr(self._form_dict, name)

    @property
    def form_dict(self):
        """The Form dictionary of this Form."""

        return self._form_dict

    @property
    def interaction(self):
        """Interaction that contains this Form."""

        return self._interaction

    @property
    def protocol(self):
        """Form protocol."""

        return self._protocol

    @property
    def id(self):
        """Returns the ID of this Form.
        The ID is a hash that is based on the Form attributes.
        No two Forms with the same ID may exist within the same Interaction.
        The ID of a Form could change during its lifetime if some attributes are updated."""

        return hash((
            self.protocol,
            self.href,
            self.content_type,
            self.op
        ))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import uuid

import pytest
import tornado.gen
import tornado.ioloop
import tornado.web
from faker import Faker
from mock import MagicMock

from tests.td_examples import TD_EXAMPLE
from tests.utils import find_free_port
from wotpy.protocols.client import BaseProtocolClient
from wotpy.wot.consumed.thing import ConsumedThing
from wotpy.wot.dictionaries.interaction import PropertyFragmentDict, ActionFragmentDict, EventFragmentDict
from wotpy.wot.exposed.thing import ExposedThing
from wotpy.wot.servient import Servient
from wotpy.wot.td import ThingDescription
from wotpy.wot.thing import Thing


def _build_property_fragment():
    """Builds and returns a random Property init fragment."""

    return PropertyFragmentDict({
        "description": Faker().sentence(),
        "readOnly": False,
        "observable": True,
        "type": "string"
    })


def _build_event_fragment():
    """Builds and returns a random Event init fragment."""

    return EventFragmentDict({
        "description": Faker().sentence(),
        "data": {"type": "string"}
    })


def _build_action_fragment():
    """Builds and returns a random Action init fragment."""

    return ActionFragmentDict({
        "description": Faker().sentence(),
        "input": {
            "type": "string",
            "description": Faker().sentence()
        },
        "output": {
            "type": "string",
            "description": Faker().sentence()
        }
    })


@pytest.fixture
def property_fragment():
    """Builds and returns a random Property init fragment."""

    return _build_property_fragment()


@pytest.fixture
def action_fragment():
    """Builds and returns a random ActionInit."""

    return _build_action_fragment()


@pytest.fixture
def event_fragment():
    """Builds and returns a random EventInit."""

    return _build_event_fragment()


@pytest.fixture
def exposed_thing():
    """Builds and returns a random ExposedThing."""

    return ExposedThing(
        servient=Servient(),
        thing=Thing(id=uuid.uuid4().urn))


@pytest.fixture
def td_example_tornado_app():
    """Builds a Tornado web application with a simple handler
    that exposes the example Thing Description document."""

    # noinspection PyAbstractClass
    class TDHandler(tornado.web.RequestHandler):
        """Dummy handler to fetch a JSON-serialized TD document."""

        def get(self):
            self.write(TD_EXAMPLE)

    return tornado.web.Application([(r"/", TDHandler)])


class ExposedThingProxyClient(BaseProtocolClient):
    """Dummy Protocol Binding client implementation that
    basically serves as a proxy for a local ExposedThing object."""

    def __init__(self, exp_thing):
        self._exp_thing = exp_thing
        super(ExposedThingProxyClient, self).__init__()

    @property
    def protocol(self):
        return None

    def is_supported_interaction(self, td, name):
        return True

    @tornado.gen.coroutine
    def invoke_action(self, td, name, input_value, timeout=None):
        result = yield self._exp_thing.invoke_action(name, input_value)
        raise tornado.gen.Return(result)

    @tornado.gen.coroutine
    def write_property(self, td, name, value, timeout=None):
        yield self._exp_thing.write_property(name, value)

    @tornado.gen.coroutine
    def read_property(self, td, name, timeout=None):
        value = yield self._exp_thing.read_property(name)
        raise tornado.gen.Return(value)

    def on_event(self, td, name):
        return self._exp_thing.on_event(name)

    def on_property_change(self, td, name):
        return self._exp_thing.on_property_change(name)

    def on_td_change(self, url):
        return self._exp_thing.on_td_change()


@pytest.fixture
def consumed_exposed_pair():
    """Returns a dict with two keys:
    * consumed_thing: A ConsumedThing instance. The Servient instance that contains this
    ConsumedThing has been patched to use the ExposedThingProxyClient Protocol Binding client.
    * exposed_thing: The ExposedThing behind the previous ConsumedThing (for assertion purposes)."""

    servient = Servient()

    exp_thing = ExposedThing(
        servient=servient,
        thing=Thing(id=uuid.uuid4().urn))

    servient.select_client = MagicMock(return_value=ExposedThingProxyClient(exp_thing))

    @tornado.gen.coroutine
    def lower(parameters):
        input_value = parameters.get("input")
        yield tornado.gen.sleep(0)
        raise tornado.gen.Return(str(input_value).lower())

    exp_thing.add_property(uuid.uuid4().hex, _build_property_fragment())
    exp_thing.add_action(uuid.uuid4().hex, _build_action_fragment(), lower)
    exp_thing.add_event(uuid.uuid4().hex, _build_event_fragment())

    td = ThingDescription.from_thing(exp_thing.thing)

    return {
        "consumed_thing": ConsumedThing(servient=servient, td=td),
        "exposed_thing": exp_thing
    }


@pytest.fixture(params=[{"catalogue_enabled": True}])
def servient(request):
    """Returns an empty WoT Servient."""

    catalogue_port = find_free_port() if request.param.get('catalogue_enabled') else None

    servient = Servient(catalogue_port=catalogue_port)

    @tornado.gen.coroutine
    def start():
        yield servient.start()

    tornado.ioloop.IOLoop.current().run_sync(start)

    yield servient

    @tornado.gen.coroutine
    def shutdown():
        yield servient.shutdown()

    tornado.ioloop.IOLoop.current().run_sync(shutdown)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
HTTP request handlers that implement each of the Interaction verbs.

.. autosummary::
    :toctree: _handlers

    wotpy.protocols.http.handlers.action
    wotpy.protocols.http.handlers.event
    wotpy.protocols.http.handlers.property
    wotpy.protocols.http.handlers.utils
"""



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import time
import uuid
# noinspection PyCompatibility
from concurrent.futures import ThreadPoolExecutor

import pytest
import six
import tornado.gen
import tornado.ioloop
from faker import Faker
from slugify import slugify
from tornado.concurrent import Future

from tests.utils import run_test_coroutine
from wotpy.wot.dictionaries.interaction import PropertyFragmentDict
from wotpy.wot.dictionaries.thing import ThingFragment
from wotpy.wot.enums import TDChangeMethod, TDChangeType, DataType
from wotpy.wot.exposed.thing import ExposedThing
from wotpy.wot.servient import Servient
from wotpy.wot.thing import Thing


def _test_td_change_events(exposed_thing, property_fragment, event_fragment, action_fragment, subscribe_func):
    """Helper function to test subscriptions to TD changes."""

    @tornado.gen.coroutine
    def test_coroutine():
        prop_name = Faker().pystr()
        event_name = Faker().pystr()
        action_name = Faker().pystr()

        complete_futures = {
            (TDChangeType.PROPERTY, TDChangeMethod.ADD): Future(),
            (TDChangeType.PROPERTY, TDChangeMethod.REMOVE): Future(),
            (TDChangeType.EVENT, TDChangeMethod.ADD): Future(),
            (TDChangeType.EVENT, TDChangeMethod.REMOVE): Future(),
            (TDChangeType.ACTION, TDChangeMethod.ADD): Future(),
            (TDChangeType.ACTION, TDChangeMethod.REMOVE): Future()
        }

        def on_next(ev):
            change_type = ev.data.td_change_type
            change_method = ev.data.method
            interaction_name = ev.data.name
            future_key = (change_type, change_method)
            complete_futures[future_key].set_result(interaction_name)

        subscription = subscribe_func(on_next)

        yield tornado.gen.sleep(0)

        exposed_thing.add_event(event_name, event_fragment)

        assert complete_futures[(TDChangeType.EVENT, TDChangeMethod.ADD)].result() == event_name
        assert not complete_futures[(TDChangeType.EVENT, TDChangeMethod.REMOVE)].done()

        exposed_thing.remove_event(name=event_name)
        exposed_thing.add_property(prop_name, property_fragment)

        assert complete_futures[(TDChangeType.EVENT, TDChangeMethod.REMOVE)].result() == event_name
        assert complete_futures[(TDChangeType.PROPERTY, TDChangeMethod.ADD)].result() == prop_name
        assert not complete_futures[(TDChangeType.PROPERTY, TDChangeMethod.REMOVE)].done()

        exposed_thing.remove_property(name=prop_name)
        exposed_thing.add_action(action_name, action_fragment)
        exposed_thing.remove_action(name=action_name)

        assert complete_futures[(TDChangeType.PROPERTY, TDChangeMethod.REMOVE)].result() == prop_name
        assert complete_futures[(TDChangeType.ACTION, TDChangeMethod.ADD)].result() == action_name
        assert complete_futures[(TDChangeType.ACTION, TDChangeMethod.REMOVE)].result() == action_name

        subscription.dispose()

    run_test_coroutine(test_coroutine)


def test_thing_template_getters(exposed_thing):
    """ThingTemplate properties can be accessed from the ExposedThing."""

    thing_template = exposed_thing.thing.thing_fragment

    assert exposed_thing.id == thing_template.id
    assert exposed_thing.name == thing_template.name
    assert exposed_thing.description == thing_template.description


def test_read_property(exposed_thing, property_fragment):
    """Properties may be retrieved on ExposedThings."""

    @tornado.gen.coroutine
    def test_coroutine():
        prop_name = Faker().pystr()
        prop_init_value = Faker().sentence()
        exposed_thing.add_property(prop_name, property_fragment, value=prop_init_value)
        value = yield exposed_thing.read_property(prop_name)
        assert value == prop_init_value

    run_test_coroutine(test_coroutine)


def test_write_property(exposed_thing, property_fragment):
    """Properties may be updated on ExposedThings."""

    assert property_fragment.writable

    @tornado.gen.coroutine
    def test_coroutine():
        updated_val = Faker().pystr()
        prop_name = Faker().pystr()

        exposed_thing.add_property(prop_name, property_fragment)

        yield exposed_thing.write_property(prop_name, updated_val)

        value = yield exposed_thing.read_property(prop_name)

        assert value == updated_val

    run_test_coroutine(test_coroutine)


def test_write_non_writable_property(exposed_thing):
    """Attempts to write a non-writable property should return an error."""

    prop_init_non_writable = PropertyFragmentDict({
        "type": "string",
        "readOnly": True
    })

    @tornado.gen.coroutine
    def test_coroutine():
        prop_name = Faker().pystr()
        exposed_thing.add_property(prop_name, prop_init_non_writable)

        with pytest.raises(Exception):
            yield exposed_thing.write_property(prop_name, Faker().pystr())

    run_test_coroutine(test_coroutine)


def test_invoke_action(exposed_thing, action_fragment):
    """Actions can be invoked on ExposedThings."""

    thread_executor = ThreadPoolExecutor(max_workers=1)

    def upper_thread(parameters):
        input_value = parameters.get("input")
        return thread_executor.submit(lambda x: time.sleep(0.1) or str(x).upper(), input_value)

    def upper(parameters):
        loop = tornado.ioloop.IOLoop.current()
        input_value = parameters.get("input")
        return loop.run_in_executor(None, lambda x: time.sleep(0.1) or str(x).upper(), input_value)

    @tornado.gen.coroutine
    def lower(parameters):
        input_value = parameters.get("input")
        yield tornado.gen.sleep(0)
        raise tornado.gen.Return(str(input_value).lower())

    def title(parameters):
        input_value = parameters.get("input")
        future = Future()
        future.set_result(input_value.title())
        return future

    handlers_map = {
        upper_thread: lambda x: x.upper(),
        upper: lambda x: x.upper(),
        lower: lambda x: x.lower(),
        title: lambda x: x.title()
    }

    @tornado.gen.coroutine
    def test_coroutine():
        action_name = Faker().pystr()
        exposed_thing.add_action(action_name, action_fragment)

        for handler, assert_func in six.iteritems(handlers_map):
            exposed_thing.set_action_handler(action_name, handler)
            action_arg = Faker().sentence(10)
            result = yield exposed_thing.invoke_action(action_name, action_arg)
            assert result == assert_func(action_arg)

    run_test_coroutine(test_coroutine)


def test_invoke_action_undefined_handler(exposed_thing, action_fragment):
    """Actions with undefined handlers return an error."""

    @tornado.gen.coroutine
    def test_coroutine():
        action_name = Faker().pystr()
        exposed_thing.add_action(action_name, action_fragment)

        with pytest.raises(NotImplementedError):
            yield exposed_thing.invoke_action(action_name)

        @tornado.gen.coroutine
        def dummy_func(parameters):
            assert parameters.get("input") is None
            raise tornado.gen.Return(True)

        exposed_thing.set_action_handler(action_name, dummy_func)

        result = yield exposed_thing.invoke_action(action_name)

        assert result

    run_test_coroutine(test_coroutine)


def test_on_property_change(exposed_thing, property_fragment):
    """Property changes can be observed."""

    assert property_fragment.observable

    @tornado.gen.coroutine
    def test_coroutine():
        prop_name = Faker().pystr()
        exposed_thing.add_property(prop_name, property_fragment)

        observable_prop = exposed_thing.on_property_change(prop_name)

        property_values = Faker().pylist(5, True, *(str,))

        emitted_values = []

        def on_next_property_event(ev):
            emitted_values.append(ev.data.value)

        subscription = observable_prop.subscribe(on_next_property_event)

        for val in property_values:
            yield exposed_thing.write_property(prop_name, val)

        assert emitted_values == property_values

        subscription.dispose()

    run_test_coroutine(test_coroutine)


def test_on_property_change_non_observable(exposed_thing):
    """Observe requests to non-observable properties are rejected."""

    prop_init_non_observable = PropertyFragmentDict({
        "type": "string",
        "observable": False
    })

    @tornado.gen.coroutine
    def test_coroutine():
        prop_name = Faker().pystr()
        exposed_thing.add_property(prop_name, prop_init_non_observable)

        observable_prop = exposed_thing.on_property_change(prop_name)

        future_next = Future()
        future_error = Future()

        def on_next(item):
            future_next.set_result(item)

        def on_error(err):
            future_error.set_exception(err)

        subscription = observable_prop.subscribe(on_next=on_next, on_error=on_error)

        yield exposed_thing.write_property(prop_name, Faker().pystr())

        with pytest.raises(Exception):
            future_error.result()

        assert not future_next.done()

        subscription.dispose()

    run_test_coroutine(test_coroutine)


def test_on_event(exposed_thing, event_fragment):
    """Events defined in the Thing Description can be observed."""

    event_name = Faker().pystr()
    exposed_thing.add_event(event_name, event_fragment)

    observable_event = exposed_thing.on_event(event_name)

    event_payloads = [Faker().pystr() for _ in range(5)]

    emitted_payloads = []

    def on_next_event(ev):
        emitted_payloads.append(ev.data)

    subscription = observable_event.subscribe(on_next_event)

    for val in event_payloads:
        exposed_thing.emit_event(event_name, val)

    assert emitted_payloads == event_payloads

    subscription.dispose()


def test_on_td_change(exposed_thing, property_fragment, event_fragment, action_fragment):
    """Thing Description changes can be observed."""

    def subscribe_func(*args, **kwargs):
        return exposed_thing.on_td_change().subscribe(*args, **kwargs)

    _test_td_change_events(exposed_thing, property_fragment, event_fragment, action_fragment, subscribe_func)


def test_thing_property_get(exposed_thing, property_fragment):
    """Property values can be retrieved on ExposedThings using the map-like interface."""

    @tornado.gen.coroutine
    def test_coroutine():
        prop_name = Faker().pystr()
        prop_init_value = Faker().sentence()
        exposed_thing.add_property(prop_name, property_fragment, value=prop_init_value)
        value = yield exposed_thing.properties[prop_name].read()
        assert value == prop_init_value

    run_test_coroutine(test_coroutine)


def test_thing_property_set(exposed_thing, property_fragment):
    """Property values can be updated on ExposedThings using the map-like interface."""

    assert property_fragment.writable

    @tornado.gen.coroutine
    def test_coroutine():
        updated_val = Faker().pystr()
        prop_name = Faker().pystr()
        exposed_thing.add_property(prop_name, property_fragment)
        yield exposed_thing.properties[prop_name].write(updated_val)
        value = yield exposed_thing.properties[prop_name].read()
        assert value == updated_val

    run_test_coroutine(test_coroutine)


def test_thing_property_subscribe(exposed_thing, property_fragment):
    """Property updates can be observed on ExposedThings using the map-like interface."""

    assert property_fragment.observable

    @tornado.gen.coroutine
    def test_coroutine():
        prop_name = Faker().pystr()
        exposed_thing.add_property(prop_name, property_fragment)

        values = [Faker().sentence() for _ in range(10)]
        values_futures = {key: Future() for key in values}

        def on_next(ev):
            value = ev.data.value
            if value in values_futures and not values_futures[value].done():
                values_futures[value].set_result(True)

        subscription = exposed_thing.properties[prop_name].subscribe(on_next)

        yield tornado.gen.sleep(0)

        for val in values:
            yield exposed_thing.properties[prop_name].write(val)

        yield [future for future in six.itervalues(values_futures)]

        subscription.dispose()

    run_test_coroutine(test_coroutine)


def test_thing_property_getters(exposed_thing, property_fragment):
    """ThingProperty retrieved from ExposedThing expose the attributes
    from the Interaction, InteractionFragment and PropertyFragment interfaces."""

    @tornado.gen.coroutine
    def test_coroutine():
        prop_name = Faker().pystr()
        prop_init_value = Faker().sentence()
        exposed_thing.add_property(prop_name, property_fragment, value=prop_init_value)
        thing_property = exposed_thing.properties[prop_name]

        assert len(thing_property.forms) == 0
        assert thing_property.title == property_fragment.title
        assert thing_property.description == property_fragment.description
        assert thing_property.observable == property_fragment.observable
        assert thing_property.type == property_fragment.type

    run_test_coroutine(test_coroutine)


def test_thing_action_run(exposed_thing, action_fragment):
    """Actions can be invoked on ExposedThings using the map-like interface."""

    @tornado.gen.coroutine
    def lower(parameters):
        input_value = parameters.get("input")
        raise tornado.gen.Return(str(input_value).lower())

    @tornado.gen.coroutine
    def test_coroutine():
        action_name = Faker().pystr()
        exposed_thing.add_action(action_name, action_fragment, lower)
        input_value = Faker().pystr()

        result = yield exposed_thing.actions[action_name].invoke(input_value)
        result_expected = yield exposed_thing.invoke_action(action_name, input_value)

        assert result == result_expected

    run_test_coroutine(test_coroutine)


def test_thing_action_getters(exposed_thing, action_fragment):
    """ThingAction retrieved from ExposedThing expose the attributes
    from the Interaction, InteractionFragment and ActionFragment interfaces."""

    @tornado.gen.coroutine
    def test_coroutine():
        action_name = Faker().pystr()
        exposed_thing.add_action(action_name, action_fragment)
        thing_action = exposed_thing.actions[action_name]

        assert len(thing_action.forms) == 0
        assert thing_action.title == action_fragment.title
        assert thing_action.description == action_fragment.description
        assert thing_action.input.type == action_fragment.input.type
        assert thing_action.output.type == action_fragment.output.type

    run_test_coroutine(test_coroutine)


def test_thing_event_subscribe(exposed_thing, event_fragment):
    """Property updates can be observed on ExposedThings using the map-like interface."""

    @tornado.gen.coroutine
    def test_coroutine():
        event_name = Faker().pystr()
        exposed_thing.add_event(event_name, event_fragment)

        values = [Faker().sentence() for _ in range(10)]
        values_futures = {key: Future() for key in values}

        def on_next(ev):
            value = ev.data
            if value in values_futures and not values_futures[value].done():
                values_futures[value].set_result(True)

        subscription = exposed_thing.events[event_name].subscribe(on_next)

        yield tornado.gen.sleep(0)

        for val in values:
            yield exposed_thing.events[event_name].emit(val)

        yield [future for future in six.itervalues(values_futures)]

        subscription.dispose()

    run_test_coroutine(test_coroutine)


def test_thing_event_getters(exposed_thing, event_fragment):
    """ThingEvent retrieved from ExposedThing expose the attributes
    from the Interaction, InteractionFragment and EventFragment interfaces."""

    @tornado.gen.coroutine
    def test_coroutine():
        event_name = Faker().pystr()
        exposed_thing.add_event(event_name, event_fragment)
        thing_event = exposed_thing.events[event_name]

        assert len(thing_event.forms) == 0
        assert thing_event.title == event_fragment.title
        assert thing_event.description == event_fragment.description
        assert thing_event.data.type == event_fragment.data.type

    run_test_coroutine(test_coroutine)


def test_set_property_read_handler(exposed_thing, property_fragment):
    """Read handlers can be defined for ExposedThing property interactions."""

    const_prop_value = Faker().sentence()

    @tornado.gen.coroutine
    def read_handler():
        raise tornado.gen.Return(const_prop_value)

    @tornado.gen.coroutine
    def test_coroutine():
        prop_name = Faker().pystr()
        exposed_thing.add_property(prop_name, property_fragment)
        exposed_thing.set_property_read_handler(prop_name, read_handler)
        value = yield exposed_thing.properties[prop_name].read()
        assert value == const_prop_value

    run_test_coroutine(test_coroutine)


def test_set_property_write_handler(exposed_thing, property_fragment):
    """Write handlers can be defined for ExposedThing property interactions."""

    prop_history = []

    @tornado.gen.coroutine
    def write_handler(value):
        yield tornado.gen.sleep(0)
        prop_history.append(value)

    @tornado.gen.coroutine
    def test_coroutine():
        prop_name = Faker().pystr()
        exposed_thing.add_property(prop_name, property_fragment)
        exposed_thing.set_property_write_handler(prop_name, write_handler)
        prop_value = Faker().sentence()
        assert not len(prop_history)
        yield exposed_thing.properties[prop_name].write(prop_value)
        assert prop_value in prop_history

    run_test_coroutine(test_coroutine)


def test_subscribe(exposed_thing, property_fragment, event_fragment, action_fragment):
    """Subscribing to the ExposedThing provides Thing Description update events."""

    def subscribe_func(*args, **kwargs):
        return exposed_thing.subscribe(*args, **kwargs)

    _test_td_change_events(exposed_thing, property_fragment, event_fragment, action_fragment, subscribe_func)


def test_thing_interaction_dict_behaviour(exposed_thing, property_fragment):
    """The Interactions dict-like interface of an ExposedThing behaves like a dict."""

    prop_name = Faker().pystr()
    exposed_thing.add_property(prop_name, property_fragment)

    assert len(exposed_thing.properties) == 1
    assert prop_name in exposed_thing.properties
    assert next(key for key in exposed_thing.properties if key == prop_name)


def test_thing_fragment_getters_setters():
    """ThingFragment attributes can be get and set from the ExposedThing."""

    thing_fragment = ThingFragment({
        "id": uuid.uuid4().urn,
        "name": Faker().pystr(),
        "description": Faker().pystr(),
        "properties": {
            uuid.uuid4().hex: {
                "description": Faker().pystr(),
                "type": DataType.STRING
            }
        }
    })

    thing = Thing(thing_fragment=thing_fragment)
    exp_thing = ExposedThing(servient=Servient(), thing=thing)

    assert exp_thing.name == thing_fragment.name
    assert exp_thing.description == thing_fragment.description
    assert list(exp_thing.properties) == list(six.iterkeys(thing_fragment.properties))

    name_original = thing_fragment.name
    name_updated = Faker().pystr()

    description_original = thing_fragment.description
    description_updated = Faker().pystr()

    exp_thing.name = name_updated
    exp_thing.description = description_updated

    assert exp_thing.name == name_updated
    assert exp_thing.name != name_original
    assert exp_thing.description == description_updated
    assert exp_thing.description != description_original

    with pytest.raises(AttributeError):
        # noinspection PyPropertyAccess
        exp_thing.id = Faker().pystr()

    with pytest.raises(AttributeError):
        # noinspection PyPropertyAccess
        exp_thing.properties = Faker().pylist()

    with pytest.raises(AttributeError):
        # noinspection PyPropertyAccess
        exp_thing.actions = Faker().pylist()

    with pytest.raises(AttributeError):
        # noinspection PyPropertyAccess
        exp_thing.events = Faker().pylist()


def _test_equivalent_interaction_names(base_name, transform_name):
    """Helper function to test that interaction names
    are equivalent given a certain transformation function."""

    thing = Thing(id=uuid.uuid4().urn)
    exp_thing = ExposedThing(servient=Servient(), thing=thing)

    prop_name = "property" + base_name
    prop_name_transform = transform_name(prop_name)

    prop_default_value = Faker().pybool()
    exp_thing.add_property(prop_name, {"type": DataType.BOOLEAN}, value=prop_default_value)

    with pytest.raises(ValueError):
        exp_thing.add_property(prop_name_transform, {"type": DataType.BOOLEAN})

    @tornado.gen.coroutine
    def assert_prop_read():
        assert (yield exp_thing.properties[prop_name].read()) is prop_default_value
        assert (yield exp_thing.properties[prop_name_transform].read()) is prop_default_value

    tornado.ioloop.IOLoop.current().run_sync(assert_prop_read)

    action_name = "action" + base_name
    action_name_transform = transform_name(action_name)

    exp_thing.add_action(action_name, {})

    with pytest.raises(ValueError):
        exp_thing.add_action(action_name_transform, {})

    assert exp_thing.actions[action_name]
    assert exp_thing.actions[action_name_transform]

    event_name = "event" + base_name
    event_name_transform = transform_name(event_name)

    exp_thing.add_event(event_name, {})

    with pytest.raises(ValueError):
        exp_thing.add_event(event_name_transform, {})

    assert exp_thing.events[event_name]
    assert exp_thing.events[event_name_transform]


def test_interaction_name_case_insensitive():
    """ExposedThing interaction names are equivalent in a case-insensitive fashion."""

    _test_equivalent_interaction_names("camelCaseStr", lambda name: name.lower())
    _test_equivalent_interaction_names("camelCaseStr", lambda name: name.upper())
    _test_equivalent_interaction_names("camelCaseStr", lambda name: name.title())


def test_interaction_name_url_safe():
    """ExposedThing interaction names are equivalent in a URL-safe fashion."""

    _test_equivalent_interaction_names("url_UnSafE-Str", lambda name: slugify(name))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
import uuid

import pytest
import tornado.gen
import tornado.ioloop
from faker import Faker

from tests.utils import find_free_port
from wotpy.support import is_coap_supported
from wotpy.wot.dictionaries.interaction import PropertyFragmentDict, EventFragmentDict, ActionFragmentDict
from wotpy.wot.exposed.thing import ExposedThing
from wotpy.wot.servient import Servient
from wotpy.wot.td import ThingDescription
from wotpy.wot.thing import Thing

collect_ignore = []

if not is_coap_supported():
    logging.warning("Skipping CoAP tests due to unsupported platform")
    collect_ignore += ["test_server.py", "test_client.py"]


@pytest.fixture(params=[{"action_clear_ms": 5000}])
def coap_server(request):
    """Builds a CoAPServer instance that contains an ExposedThing."""

    from wotpy.protocols.coap.server import CoAPServer

    exposed_thing = ExposedThing(servient=Servient(), thing=Thing(id=uuid.uuid4().urn))

    exposed_thing.add_property(uuid.uuid4().hex, PropertyFragmentDict({
        "type": "number",
        "observable": True
    }), value=Faker().pyint())

    exposed_thing.add_property(uuid.uuid4().hex, PropertyFragmentDict({
        "type": "string",
        "observable": True
    }), value=Faker().pyint())

    exposed_thing.add_event(uuid.uuid4().hex, EventFragmentDict({
        "type": "object"
    }))

    action_name = uuid.uuid4().hex

    @tornado.gen.coroutine
    def triple(parameters):
        input_value = parameters.get("input")
        raise tornado.gen.Return(input_value * 3)

    exposed_thing.add_action(action_name, ActionFragmentDict({
        "input": {"type": "number"},
        "output": {"type": "number"}
    }), triple)

    port = find_free_port()

    server = CoAPServer(port=port, **request.param)
    server.add_exposed_thing(exposed_thing)
    server.start()

    @tornado.gen.coroutine
    def start():
        yield server.start()

    tornado.ioloop.IOLoop.current().run_sync(start)

    yield server

    @tornado.gen.coroutine
    def stop():
        yield server.stop()

    tornado.ioloop.IOLoop.current().run_sync(stop)


@pytest.fixture
def coap_servient():
    """Returns a Servient that exposes a CoAP server and one ExposedThing."""

    from wotpy.protocols.coap.server import CoAPServer

    coap_port = find_free_port()
    the_coap_server = CoAPServer(port=coap_port)

    servient = Servient(catalogue_port=None)
    servient.add_server(the_coap_server)

    @tornado.gen.coroutine
    def start():
        raise tornado.gen.Return((yield servient.start()))

    wot = tornado.ioloop.IOLoop.current().run_sync(start)

    property_name_01 = uuid.uuid4().hex
    action_name_01 = uuid.uuid4().hex
    event_name_01 = uuid.uuid4().hex

    td_dict = {
        "id": uuid.uuid4().urn,
        "name": uuid.uuid4().hex,
        "properties": {
            property_name_01: {
                "observable": True,
                "type": "string"
            }
        },
        "actions": {
            action_name_01: {
                "input": {
                    "type": "number"
                },
                "output": {
                    "type": "number"
                },
            }
        },
        "events": {
            event_name_01: {
                "type": "string"
            }
        },
    }

    td = ThingDescription(td_dict)

    exposed_thing = wot.produce(td.to_str())
    exposed_thing.expose()

    @tornado.gen.coroutine
    def action_handler(parameters):
        input_value = parameters.get("input")
        raise tornado.gen.Return(int(input_value) * 2)

    exposed_thing.set_action_handler(action_name_01, action_handler)

    yield servient

    @tornado.gen.coroutine
    def shutdown():
        yield servient.shutdown()

    tornado.ioloop.IOLoop.current().run_sync(shutdown)
